from clint.textui import colored, puts, puts_err, indent
import os
from . import message


def get_genome_directory_file():
    """
        Return genome directory.
    """
    module_path = os.path.split(os.path.realpath(__file__))[0]
    return module_path + "/.genome_directory"


def get_genome_directory():
    genome_directory_file = get_genome_directory_file()
    if os.path.exists(genome_directory_file):
        with open(genome_directory_file, "r") as f:
            genome_directory = os.path.abspath(os.path.expanduser(f.read()))      
    else:
        with open(genome_directory_file, "w") as f:
            genome_directory = os.path.expanduser("~/.genome/")
            f.write(genome_directory)
    # Create directory if not exists
    if not os.path.isdir(genome_directory):
        os.mkdir(genome_directory)
    return genome_directory


def get_genome_list():
    """
        Return list of downloaded genomes.
    """
    return [os.path.split(x[0])[1] for x in os.walk(get_genome_directory())][1:]



def output_genome_list():
    """
        Outputs list of available genomes
    """
    message('\n'.join(get_genome_list()))


def resolve_reference_genome(loc):
    """
        Resolve location of reference genome file.
    """

    if loc is None:
        message("You must specify a genome:")
        output_genome_list()
        exit()

    if os.path.exists(loc):
        return loc
    else:
        if loc in get_genome_list():
            reference_location = "{gd}/{loc}/{loc}.fa.gz".format(gd = get_genome_directory(), loc = loc)
            with indent(4):
                puts_err(colored.green("\nUsing reference located at %s\n" % reference_location))
            return reference_location
        else:
            with indent(4):
                exit(puts_err(colored.red("\nGenome '%s' does not exist\n" % loc)))



#! /usr/bin/env python
"""
usage:
  vk stat <vcf>

"""
from docopt import docopt
from subprocess import Popen, PIPE
from signal import signal, SIGPIPE, SIG_DFL
from .utils import autoconvert
import re
signal(SIGPIPE, SIG_DFL)


debug = None

class out_line:
    def __init__(self):
        self.vcf = ""
        self.e1 = ""
        self.e2 = ""
        self.e3 = ""
        self.val = ""

    def __str__(self):
        return '\t'.join([self.vcf, self.e1, self.e2, self.e3, self.val])


class stat:
    """
      Class for parsing and manipulating bcf
    """


if __name__ == '__main__':
    args = docopt(__doc__,
                  version='VCF-Toolbox v0.1',
                  argv=debug)
    if args["<vcf>"] == "":
        print(__doc__)
    out = out_line()
    out.vcf = args["<vcf>"]
    comm = ["bcftools", "stats", "--depth", "0,1000,1", "-s", "-", "-v", args["<vcf>"]]
    result = Popen(comm, stdout = PIPE, stderr = PIPE)
    for n, line in enumerate(result.stdout):
        line = re.sub("\[[0-9]+\]","", line)
        line = line.strip().split("\t")
        if len(line) > 1:
            if line[0].startswith("# SN"):
                out.e1 = "Summary numbers"
            elif line[0].startswith("SN"):
                out.e2 = line[2].strip(":")
                out.val = line[3]
                print(out)
            elif line[0] in ["# TSTV", "# SiS", "# ST"]:
                variables = line[2:]
                out.e2 = line[0].strip("# ")
            elif line[0] in ["# AF", "# DP",  "# HWE", "# PSC"]:
                variables = line[2:]
                out.e2 = line[0].strip("# ")
            else:
                out.e1 = line[0]
                if line[0] in ["TSTV", "SiS"]:
                    out.e3 = ""
                    for k, v in zip(variables, line[2:]):
                        out.e2 = k
                        out.val = v
                        print(out)
                elif line[0] in ["ST"]:
                    out.e2 = line[2]
                    out.e3 = ""
                    out.val = line[3]
                    print(out)
                elif line[0] in ["AF", "QUAL", "DP", "HWE", "PSC"]:
                    out.e1 = line[0]

                    out.e3 = line[2]
                    for k,v in zip(variables, line[2:]):
                        out.e2 = k
                        if type(autoconvert(v)) != str:
                            out.val = v
                            print(out)




#! /usr/bin/env python
"""
usage:
  vk tajima [--no-header --extra] <window-size> <step-size> <vcf>
  vk tajima [--no-header --extra] <window-size> --sliding <vcf>

options:
  -h --help                   Show this screen.
  --version                   Show version.
  --window-size               Size of window
  --step-size                 Distance to move window.
  --extra                     display extra

command:
  tajima        Calculate Tajima's D


output:
    CHROM
    BIN_START
    BIN_END
    N_Sites
    N_SNPs
    TajimaD


"""
import os
from math import isinf
from signal import SIG_DFL, SIGPIPE, signal

from clint.textui import colored, puts_err
from docopt import docopt
from vcfkit import __version__
from vcfkit.utils.vcf import vcf

signal(SIGPIPE, SIG_DFL)


debug = None


class tajima(vcf):

    """
      Subclass of the vcf object
      used for calculating tajima's D
    """

    def __init__(self, filename):
        vcf.__init__(self, filename)

    def calc_tajima(self, window_size, step_size, sliding, extra=False):
        # Tajima D Constants
        n = self.n * 2
        a1 = sum([1.0 / i for i in range(1, n)])
        a2 = sum([1.0 / i**2.0 for i in range(1, n)])
        b1 = (n + 1.0) / (3.0 * (n - 1.0))
        b2 = (2.0 * (n**2 + n + 3.0)) / \
            (9.0 * n * (n - 1.0))
        c1 = b1 - (1.0 / a1)
        c2 = b2 - ((n + 2.0) / (a1 * n)) + (a2 / (a1**2))
        e1 = c1 / a1
        e2 = c2 / ((a1*a1) + a2)

        if sliding:
            shift_method = "POS-Sliding"
            step_size = None
        else:
            shift_method = "POS-Interval"

        for variant_interval in self.window(window_size=window_size, step_size=step_size, shift_method=shift_method):
            pi = 0.0
            S = 0
            n_sites = 0
            for variant in variant_interval:
                n_sites += 1
                # Use only diploid, biallelic sites, 0 < AF < 1
                if variant.ploidy == 2 and len(variant.ALT) == 1 and 0 < variant.aaf < 1 and variant.is_snp:
                    AN = variant.num_called*2
                    # j;AC : allele count in genotypes, for each ALT allele, in the same order as listed
                    AC = variant.INFO["AC"]
                    if AC > 0:
                        p = (AC*1.0/AN)*1.0
                        pi += p * (1.0-p)
                        S += 1
            try:
                CHROM = variant_interval[0].CHROM
                pi = 2.0*pi*n/(n-1.0)
                tw = (S*1.0 / a1)
                var = (e1 * S) + ((e2 * S) * (S - 1))
                TajimaD = (pi - tw) / \
                          (var)**(0.5)
                if not isinf(TajimaD) and S > 0 and AC > 0:
                    output = [CHROM,
                              variant_interval.lower_bound,
                              variant_interval.upper_bound,
                              n_sites,
                              S,
                              TajimaD]
                    if extra:
                        if step_size is None:
                            step_size = "NA"
                        output += [os.path.split(self.filename)[1],
                                   window_size,
                                   step_size]
                    yield "\t".join(map(str, output))
            except:
                pass


def large_int(i):
    return int(float(i.replace(",", "")))


def main(debug=None):
    args = docopt(__doc__,
                  version=__version__,
                  argv=debug)

    if args["<vcf>"] == "":
        print(__doc__)

    wz = large_int(args["<window-size>"])
    sz = None
    if not args["--sliding"]:
        sz = large_int(args["<step-size>"])

    if wz < sz:
        exit(puts_err(colored.red("\n\tWindow size must be >= step size.\n")))

    if args["--no-header"] is False:
        header_line = ["CHROM",
                       "BIN_START",
                       "BIN_END",
                       "N_Sites",
                       "N_SNPs",
                       "TajimaD"]
        if args["--extra"]:
            header_line += ["filename",
                            "window_size",
                            "step_size"]
        print("\t".join(header_line))
    for i in tajima(args["<vcf>"]).calc_tajima(wz, sz, args["--sliding"], extra=args["--extra"]):
        print(i)

if __name__ == '__main__':
    main()



from subprocess import check_output
from collections import OrderedDict


class sequence:
    """
        Sequence object for keeping track
        of chromosomal position.
    """
    def __init__(self, chrom, start, end, seq):
        self.chrom = chrom
        self.start = start
        self.end = end
        self.seq = seq

    def __repr__(self):
        seq_repr = ">{self.chrom}:{self.start}-{self.end}\n{self.seq}"
        return seq_repr.format(**locals())


class Fasta:
    """
        Extract regions from fasta files indexed with samtools
    """
    def __init__(self, reference):
        self.reference = reference
        self.alt_contig_names = {}

    def __getitem__(self, chrom_pos):
        if isinstance(chrom_pos, slice):
            start = chrom_pos.start
            if start < 1:
                start = 1
            end = chrom_pos.stop
            chrom_pos = "{self.chrom}:{start}-{end}".format(**locals())
            query_string = ["samtools", "faidx", self.reference, chrom_pos]
            seq = ''.join(check_output(query_string).strip().split("\n")[1:])
            seq = sequence(self.chrom_name, start, end, seq)
            return seq
        self.chrom = chrom_pos
        self.chrom_name = chrom_pos
        if chrom_pos in list(self.alt_contig_names.keys()):
            self.chrom = self.alt_contig_names[chrom_pos]
        return self

    def keys(self, weight=False):
        with open(self.reference + ".fai") as f:
            chrom_length = [x.strip().split("\t")[0:2] for x in f.readlines()]
            chrom_length = [[x[0], int(x[1])] for x in chrom_length]
            if weight is True:
                genome_length = sum([x[1] for x in chrom_length])
                return OrderedDict([x[0], 1.0*x[1]/genome_length]
                                   for x in chrom_length)
            else:
                return OrderedDict(chrom_length)



#! /usr/bin/env python
"""
usage:
  vk phylo fasta <vcf> [<region>]
  vk phylo tree (nj|upgma) [--plot] <vcf> [<region>]

options:
  -h --help                   Show this screen.
  --version                   Show version.

"""
import os
import sys
import tempfile
import webbrowser
from pkgutil import get_data
from subprocess import PIPE, Popen

import numpy as np
from jinja2 import Template

from clint.textui import colored, indent, puts_err
from docopt import docopt
from vcfkit import __version__
from vcfkit.utils import check_program_exists
from vcfkit.utils.vcf import vcf


def main(debug=None):
    args = docopt(__doc__,
                  argv=debug,
                  options_first=False,
                  version=__version__)


    def first(s):
        return s[0].replace(".", "N")

    firstv = np.vectorize(first)

    v = vcf(args["<vcf>"])

    if len(v.samples) <= 1:
        exit(puts_err(colored.red("\n\tVCF must have at least two samples.\n")))
    
    if args["<region>"]:
        variant_set = v(args["<region>"])
    else:
        variant_set = v

    if args["fasta"] or args["tree"]:
        """
            Generate an aligned fasta from a VCF file.
        """
        gt_set = np.chararray((0,len(v.samples)))
        gt_set = []
        for line in variant_set:
            if line.is_snp:
                gt_set.append(firstv(line.gt_bases))
        if len(gt_set) == 0:
            exit(puts_err("No genotypes"))
        gt_set = np.vstack(gt_set)
        seqs = list(zip(v.samples, np.transpose(gt_set)))
        if args["fasta"]:
            for sample, seq in seqs:
                print((">" + sample))
                print((''.join(seq)))

        elif args["tree"]:
            """
            Generate a phylogenetic tree using an aligned fasta with muscle.
            """

            # Check for muscle dependency
            check_program_exists("muscle")
            fasta = ""
            with indent(4):
                puts_err(colored.blue("\nGenerating Fasta\n"))
            for sample, seq in seqs:
                fasta += ">" + sample + "\n" + ''.join(seq) + "\n"
            tree_type = "upgma"  # default is upgma
            if args["nj"]:
                tree_type = "neighborjoining"
            with indent(4):
                puts_err(colored.blue("\nGenerating " + tree_type + " Tree\n"))
            comm = ["muscle", "-maketree", "-in", "-", "-cluster", tree_type]
            tree, err = Popen(comm, stdin=PIPE, stdout=PIPE).communicate(input=fasta.encode())
            
            # output tree
            print(tree.decode("utf-8"))
            
            if args["--plot"]:
                prefix = os.path.dirname(os.path.abspath(sys.modules['vcfkit'].__file__)) + "/static"
                template = open(prefix + "/tree.html",'r').read()
                tree_template = Template(template)
                html_out = tempfile.NamedTemporaryFile(suffix=".html", delete=False)
                with html_out as f:
                    tree = tree.replace("\n", "")
                    sample_len = len(v.samples)
                    f.write(tree_template.render(**locals()))
                    webbrowser.open("file://" + html_out.name)

if __name__ == '__main__':
    main()



import io
from subprocess import Popen, PIPE
import sys
import signal
from vcfkit import vk
from vcfkit.utils import *
from vcfkit.utils.blastn import blast
from Bio.Seq import Seq
from Bio.Restriction import AllEnzymes, CommOnly, RestrictionBatch
from clint.textui import colored, puts_err
signal.signal(signal.SIGINT, lambda x,y: sys.exit(0))
primer3_config =  os.path.dirname(os.path.abspath(sys.modules['vcfkit'].__file__)) + "/static/primer3_config/"

class seqprimer:
    """
        An individual primer.
    """
    def __init__(self, primer_values, template, reference, left_primer = True):
        for k,v in list(primer_values.items()):
            setattr(self, k, v)
        self.left = left_primer
        template = str(template)
        if left_primer:
            self.START = template.find(self.SEQUENCE)
        else:
            # Reverse - complement right primer to find its location
            pright_rc = str(Seq(self.SEQUENCE).reverse_complement())
            self.START = template.find(pright_rc)
        self.END = self.START + len(self.SEQUENCE)

        # Blast primer sequence
        b = blast(reference, num_alignments = 10, word_size = 14)
        self.unique_copies = b.check_primer(self.SEQUENCE)

    def filter_unique(self, copies = 1):
        return self.unique_copies <= copies

    def __str__(self):
        return self.SEQUENCE

    def __repr__(self):
        return "<primer:{self.SEQUENCE} - {self.START}-{self.END} ({self.unique_copies})>".format(**locals())


class primer_group:
    """
        A group of primers with a designated purpose.
        Can be a single primer or a pair.

        Used to calculate PCR product, start site, etc. 
    """
    def __init__(self, primer_values, template, reference, chrom, region_start):
        self.reference = reference
        self.region_start = region_start
        pleft = {k.replace("PRIMER_LEFT_", ""):v for k,v in list(primer_values.items()) if k.startswith("PRIMER_LEFT")}
        pright = {k.replace("PRIMER_RIGHT_", ""):v for k,v in list(primer_values.items()) if k.startswith("PRIMER_RIGHT")}
        self.primer_left = seqprimer(pleft, template, reference)
        self.primer_right = seqprimer(pright, template, reference, False)
        self.primer_tm = "{self.primer_left.TM},{self.primer_right.TM}".format(**locals())
        self.amplicon = template[self.primer_left.START:self.primer_right.END]
        self.amp_start = self.region_start + self.primer_left.START
        self.amp_end = self.region_start + self.primer_right.END
        self.amplicon_length = len(self.amplicon)
        self.amplicon_region = "{chrom}:{self.amp_start}-{self.amp_end}".format(**locals())
        for k,v in list(primer_values.items()):
            if not k.startswith("PRIMER_LEFT") and not k.startswith("PRIMER_RIGHT"):
                setattr(self, k, v)

    def filter_primer_group(self):
        """
            Return a set of filtered primers that are usable.
        """
        return any([self.primer_left.filter_unique(1), self.primer_right.filter_unique(1)])

    def __repr__(self):
        return repr(self.primer_left) + "\t" + repr(self.primer_right)


class primer3:
    """
        A wrapper for primer3
    """
    def __init__(self, reference, method="pcr"):
        self.reference = reference
        self.PRIMER_GC_CLAMP = 1
        self.PRIMER_OPT_SIZE = 20
        self.PRIMER_MIN_SIZE = 18  # Must be set
        self.PRIMER_MAX_SIZE = 22  # Must be set
        self.PRIMER_NUM_RETURN = 5
        self.PRIMER_PRODUCT_SIZE_RANGE = "200-1500"
        
        if method == "pcr":
            seq_template_length = self.PRIMER_PRODUCT_SIZE_RANGE.split("-")[1]
            self.seq_template_length = int(seq_template_length)
            self.PRIMER_TASK = "pick_pcr_primers"
            self.generate_pcr_template = True

        # Global default
        thermo_paths = ["/usr/local/share/primer3_config/",
                        "/usr/local/share/primer3/primer3_config/",
                        "~/.linuxbrew/share/primer3_config/",
                        "~/.linuxbrew/share/primer3/primer3_config/",
                        "/.linuxbrew/share/primer3_config/",
                        "/.linuxbrew/share/primer3/primer3_config/",
                        primer3_config]
        paths = [x for x in thermo_paths if os.path.exists(os.path.expanduser(x))]
        if len(paths) == 0:
            with indent(4):
                exit(puts_err(colored.red("\nCannot find thermo path '/primer3_config/\n")))
        thermo_path = paths[0]
        self.PRIMER_THERMODYNAMIC_PARAMETERS_PATH = thermo_path

        # Set primer returns:
        if self.PRIMER_TASK == "pick_left_only":
            self.left_or_right = ["PRIMER_LEFT"]
        else:
            self.left_or_right = ["PRIMER_LEFT", "PRIMER_RIGHT"]

    def _generate_record(self):
        # Generates text record ready for input
        # into primer3
        attributes = [x for x in dir(self) if x.upper() == x and not x.startswith("_")]
        values = [str(getattr(self, x)) for x in attributes]
        att_val = list(zip(attributes, values))
        return '\n'.join(["=".join(x) for x in att_val]) + "\n=\n"

    def fetch_primers(self, sequence_template, chrom, region_start):
        # Runs primer3 with the generated record.
        self.SEQUENCE_TEMPLATE = sequence_template
        input_record = self._generate_record().encode('utf-8')
        primer3_run = Popen(["primer3_core"], stdin=PIPE, stdout=PIPE)
        resp, err = primer3_run.communicate(input=input_record)
        resp = resp.decode().strip().split("\n")
        if err:
            exit(message(err))
        p3_results = dict([x.split("=") for x in resp
                               if x.split("=")[0] != ""])
        p3_results = {k: autoconvert(v) for k,v in list(p3_results.items())}


        if "PRIMER_LEFT_NUM_RETURNED" in p3_results:
            n_primers = p3_results["PRIMER_LEFT_NUM_RETURNED"]
            primer_return = []
            for primer_num in range(0, n_primers):
                pn = "_{n}_".format(n=primer_num)
                primer_dataset = {k.replace(pn,"_"):v for k,v in list(p3_results.items()) if pn in k}
                primer_return.append(primer_group(primer_dataset,
                                                  self.SEQUENCE_TEMPLATE,
                                                  self.reference,
                                                  chrom,
                                                  region_start))
            return primer_return



#! /usr/bin/env python
"""
usage:
  vk geno transfer-filter <vcf>
  vk geno het-polarization <vcf>


tb.py geno transfer-filter

options:
  -h --help                   Show this screen.
  --version                   Show version.



"""
from .vk import __version__
from docopt import docopt
from vcfkit.utils.vcf import *
from vcfkit.utils.fasta import *
import sys
import math
from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE, SIG_DFL)
np.set_printoptions(threshold=sys.maxsize)


def phred2p(phred):
    return 10**(phred / -10.0)


def GL2PL(gl):
    """ Converts Genotype likelyhoods to phred scaled (PL) genotype likelyhoods. """
    return -int(gl * 10)

debug = None
if len(sys.argv) == 1:
    debug = ['primer', "--ref=WBcel235", "test.vcf.gz"]


if __name__ == '__main__':
    # print debug
    args = docopt(__doc__,
                  version='VCF-Toolbox v0.1',
                  argv=debug,
                  options_first=False)
    # Locate Reference
    v = vcf(args["<vcf>"])
    format_added = False
    if args["transfer-filter"]:
        for line in v.output_raw():
            line = line.strip()
            if line.startswith("#CHROM"):
                # Get Sample information and count
                samples = line.strip().split("\t")[9:]
            elif line.startswith("#"):
                # Add Info line for het polarization flag
                if line.startswith("##FORMAT") and format_added is False:
                    format_added = True
                    line = line + "\n##FORMAT=<ID=FT,Number=1,Type=String,Description=\"Genotype-level filter\">"
            else:
                line = line.split("\t")
                FILTER = line[6]
                line[8] = line[8] + ":" + "FT"
                line[9:] = [x + ":" + FILTER for x in line[9:]]
                line = '\t'.join(line)
            sys.stdout.write(line + "\n")
    elif args["het-polarization"]:
        for line in v.output_raw():
            line = line.strip()
            if line.startswith("#CHROM"):
                # Get Sample information and count
                line = "##VCF-kit(version {v}) command=geno het-polarization\n".format(v = __version__) + line
                samples = line.split("\t")[9:]
            elif line.startswith("##FORMAT=<ID=GL,"):
                line = """##FORMAT=<ID=PL,Number=G,Type=Integer,Description="List of Phred-scaled genotype likelihoods">"""
            elif line.startswith("#"):
                # Add Info line for het polarization flag
                if line.startswith("##FORMAT") and format_added is False:
                    format_added = True
                    line += "\n##FORMAT=<ID=HP,Number=1,Type=String,Description=\"Flag used to mark whether a variant was polarized\">"
                # Pass comment lines.
            else:
                if line.split('\t')[8].find("GL") > 0:
                    # Replace GL with PL scores.
                    line = line.split('\t')
                    GL_loc = line[8].split(":").index("GL")
                    line[8] = line[8].replace("GL", "PL")
                    geno_set = []
                    for k, v in enumerate(line[9:]):
                        FT = v.split(":")
                        GL_set = FT[GL_loc].split(",")
                        try:
                            FT[GL_loc] = ','.join([str(GL2PL(float(i))) for i in GL_set])
                        except:
                            FT[GL_loc] = ",".join(GL_set)
                        geno_set.append(':'.join(FT))
                        line = line[0:9] + geno_set
                    line = '\t'.join(l)
                if line.find("0/1") > 0:
                    line = line.strip().split("\t")
                    if line[8].find("PL") > -1 and len(line[4].split(",")) == 1 and line[8].find("HP") == -1:
                        PL = line[8].split(":").index("PL")
                        add_HP_flag = 0
                        for k, v in enumerate(line[9:]):
                            PL_set = v.split(":")[PL].split(",")
                            v = v.split(":")
                            if len(PL_set) == 3:
                                PL_set = [phred2p(int(i)) for i in PL_set]
                                log_score = -math.log10(PL_set[0] / PL_set[2])
                                if add_HP_flag == 0:
                                    if line[8].find("HP") == -1:
                                        line[8] = line[8] + ":HP"
                                    add_HP_flag = 1
                                if (log_score < -2):
                                    v[0] = "0/0"
                                    if v[-1] not in ["AA", "AB", "BB"]:
                                        line[k + 9] = v + ["AA"]
                                    else:
                                        line[k + 9] = v[0:-1] + ["AA"]
                                elif (log_score > 2):
                                    v[0] = "1/1"
                                    if v[-1] not in ["AA", "AB", "BB"]:
                                        line[k + 9] = v + ["BB"]
                                    else:
                                        line[k + 9] = v[0:-1] + ["BB"]
                                else:
                                    v[0] = "0/1"
                                    if v[-1] not in ["AA", "AB", "BB"]:
                                        line[k + 9] = v + ["AB"]
                                    else:
                                        line[k + 9] = v[0:-1] + ["AB"]
                            else:
                                if add_HP_flag == 0:
                                    line[k + 9] = v
                                else:
                                    line[k + 9] = v + ["."]
                            line[k + 9] = ":".join(line[k + 9])
                    line = "\t".join(line)
            print(line)



#! /usr/bin/env python
"""
usage:
  vk genome location [<path>]
  vk genome list
  vk genome --search=<term>
  vk genome ncbi [options] --ref=<asm_name> [--accession-chrom-names]
  vk genome wormbase [options] --ref=<asm_name>

options:
  -h --help                   Show this screen.
  --directory=<dir>           Set Genome Directory

"""
from docopt import docopt
from vcfkit.utils import which, run_command
from vcfkit.utils.vcf import *
from vcfkit.utils.reference import *
from clint.textui import colored, puts, puts_err, indent, progress
import gzip
from subprocess import call
from time import time
from tabulate import tabulate as tab
import requests
import os
import urllib.request, urllib.parse, urllib.error
from sys import exit  # Used by exit(); don't remove.
from Bio import Entrez



def fetch_chrom_name(id):
    try:
        if not id.startswith("NC_"):
            return id
        Entrez.email = "vcf-kit@vcf-kit.com"
        chrom = Entrez.read(Entrez.efetch(db="nuccore", id=id, rettype="gb", retmode="xml"))
        gb_feature_quals = chrom[0]["GBSeq_feature-table"][0]["GBFeature_quals"]
        features = dict([list(x.values()) for x in gb_feature_quals])
        if "organelle" in features:
            if features["organelle"] == "mitochondrion":
                return "MtDNA"
        else:
            chrom_name = features["chromosome"]
            return chrom_name
    except:
        return id


def download_genomes(genome_db):
    if os.path.isfile(genome_db):
        fileTime = os.path.getmtime(genome_db)
    else:
        fileTime = 0
    if (time() - fileTime) > (3 * 30 * 24 * 60 * 60) or is_non_zero_file(genome_db) is False:
        with indent(2):
            puts(colored.blue('\nDownloading list of reference genomes\n'))
        r = requests.get("https://ftp.ncbi.nlm.nih.gov/genomes/ASSEMBLY_REPORTS/assembly_summary_refseq.txt", allow_redirects=True)
        genome_file = open(genome_db, "w")
        with genome_file as f:
            f.write(r.text.strip())


def is_non_zero_file(fpath):
    return os.path.isfile(fpath) and os.path.getsize(fpath) > 0


def main(debug=None):
    args = docopt(__doc__,
                  version='VCF-Toolbox v0.1',
                  argv=debug,
                  options_first=False)
    # Setup Genomes Directory
    if args["location"] and args["<path>"]:
        if args["<path>"] == "-":
            genome_directory = get_genome_directory_file()
            os.remove(genome_directory)
            return get_genome_directory_file()
        else:
            with open(get_genome_directory_file(), "w") as f:
                genome_directory = os.path.realpath(args["<path>"])
                with indent(2):
                    puts(colored.blue("\nSet genome location to: " + genome_directory + "/\n"))
                f.write(genome_directory)
                # create directory if not exists
                if not os.path.exists(genome_directory):
                    os.makedirs(genome_directory)
                return genome_directory

    if args["--directory"]:
        genome_directory = os.path.realpath(args["--ref"])
    else:
        genome_directory = get_genome_directory()

    with indent(2):
        puts(genome_directory)

    if args["location"] and not args["<path>"]:
        return genome_directory

    genome_db = get_genome_directory() + "/genomes.db"

    ################
    # List Genomes #
    ################
    if args["list"]:
        output_genome_list()

    ##################
    # Search Genomes #
    ##################
    elif args["--search"]:
        # Download and cache a list of genomes from NCBI for searching
        download_genomes(genome_db)

        # Cache result
        header = ["assembly_accession",  # 0
                  "bioproject",  # 1
                  "organism_name",  # 7
                  "asm_name",  # 15
                  "ftp_path"]  # 19

        with indent(2):
            puts(colored.blue('\nSearching...\n'))

        with open(genome_db, "r") as f:
            results = []
            for line in f:
                if not line.startswith("#"):
                    line = line.strip().split("\t")
                    line = [x for k, x in enumerate(line) if k in [0, 1, 7, 15, 19]]
                    if args["--search"].lower() in line[2].lower() and line[4] != "na":
                        results.append(line)
        with indent(4):
            puts(tab(results, headers=header))
        with indent(2):
            puts(colored.blue('\nTo download a genome and setup for use:'))
        with indent(4):
            puts(colored.green("\nvk genome ncbi --ref=<asm_name>\n"))
        return results
    elif args["--ref"]:
        # reference name.
        reference_name = args["--ref"]

        # Ensure genome db is available
        download_genomes(genome_db)

        # reference directory
        if not args["--directory"]:
            reference_directory = genome_directory + "/" + reference_name + "/"
        else:
            reference_directory = genome_directory + "/"
        if not os.path.exists(reference_directory):
            os.makedirs(reference_directory)

        # base reference filename.
        ref_filename = reference_directory + reference_name + ".tmp.fa.gz"

        if args["wormbase"]:
            asm_name = args["--ref"]
            asm_url = f"ftp://ftp.wormbase.org/pub/wormbase/releases/{asm_name}/species/c_elegans/PRJNA13758/c_elegans.PRJNA13758.{asm_name}.genomic.fa.gz"
            comm = f"curl {asm_url} > {ref_filename}"
            err = run_command(comm)
            if err != 0:
                raise Exception(colored.red(f"Wormbase genome {args['--ref']} not found."))
            # Unzip wormbase genome
            run_command(f"gzip -df {ref_filename}", shell=True)
        else:
            # NCBI
            with open(genome_db, "r") as f:
                results = []
                for line in f:
                    if not line.startswith("#"):
                        line = line.strip().split("\t")
                        line = [x for k, x in enumerate(line) if k in [0, 1, 7, 15, 19]]
                        if args["--ref"] == line[3]:
                            results.append(line)
                            reference_download = results[0]
                            url = reference_download[4].replace("ftp://", "http://") + "/" + os.path.split(reference_download[4])[1] + "_genomic.fna.gz"
            if len(results) == 0:
                with indent(2):
                    puts(colored.red('\nError: Genome ' + args["--ref"] + ' not found\n'))

            with indent(2):
                puts(colored.green('\nDownloading: ' + reference_name + "; " + url + '\n'))

            # stack overflow: 15644964;
            r = requests.get(url, stream=True)

            with open(ref_filename, 'wb') as f:
                total_length = int(r.headers.get('content-length'))
                for chunk in progress.bar(r.iter_content(chunk_size=1024), expected_size=(total_length / 1024) + 1):
                    if chunk:
                        f.write(chunk)
                        f.flush()

        # Fix chromosome names
        if not args["--accession-chrom-names"] and not args['wormbase']:
            with indent(2):
                puts(colored.green('\nFixing chromosome names\n'))

            with open(ref_filename.replace(".fa.gz", ".fa"), 'w') as outfa:
                with gzip.open(ref_filename, 'rb') as f:
                    for line in f:
                        line = line.decode("utf-8")
                        outline = line
                        if line.startswith(">"):
                            acc = line.split(" ")[0].strip(">")
                            chrom_name = fetch_chrom_name(acc)
                            if chrom_name is not None:
                                outline = ">" + chrom_name + "\n"
                            elif line.lower().find("mitochon") > 0:
                                outline = ">MtDNA\n"
                            puts(colored.blue(line.strip("\n>")) + " --> " + colored.blue(outline.strip("\n>")))
                        outfa.write(outline)

        if which("bgzip"):
            with indent(2):
                puts(colored.green('\nSwitching from gzip to bgzip\n'))
            # Convert to bgzip
            if args["--accession-chrom-names"]:
                run_command(f"gzip -df {ref_filename}", shell=True)
            comm_bgzip = "bgzip -fc {ref_filename} > {ref_out}"
            comm_bgzip = comm_bgzip.format(ref_filename=ref_filename.replace(".fa.gz", ".fa"),
                                           ref_out=ref_filename.replace(".tmp", ""))
            run_command(comm_bgzip, shell=True)
            ref_filename = ref_filename.replace(".tmp", "")
        else:
            with indent(2):
                puts_err(colored.red("Please install bgzip."))
            exit()

        if which("bwa"):
            with indent(2):
                puts(colored.green("\nCreating bwa index\n"))
            comm = f"bwa index {ref_filename}"
            run_command(comm, shell = True)
        else:
            with indent(2):
                puts(colored.blue("\nSkipping bwa index; bwa not installed\n"))

        if which("samtools"):
            with indent(2):
                puts(colored.green("\nCreating samtools index\n"))
            comm = f"samtools faidx {ref_filename}"
            run_command(comm, shell=True)
        else:
            with indent(2):
                puts(colored.blue("\nSkipping samtools index; Samtools not installed\n"))

        if which("makeblastdb") and which("gzip"):
            with indent(2):
                puts(colored.green("\nCreating blast database\n"))
            comm = "gzip -dc {ref} | makeblastdb -in - -dbtype=nucl -title={ref} -out={ref}".format(ref=ref_filename)
            run_command(comm, shell=True)
        else:
            with indent(2):
                puts(colored.blue("\nSkipping creation of blast database; blast is not installed\n"))

        # Remove temp files
        if args["--accession-chrom-names"]:
            os.remove(ref_filename.replace(".fa.gz", ".tmp.fa.gz"))

        # Remove temporary files
        try:
            os.remove(ref_filename.replace(".fa.gz", ".tmp.fa.gz"))
            os.remove(ref_filename.replace(".fa.gz", ".tmp.fa"))
        except:
            pass

        # Add error checking here...

        with indent(2):
            puts(colored.green("\nComplete!\n"))


if __name__ == '__main__':
    main()



#! /usr/bin/env python
"""
usage:
  vk hmm [options] --A=<parent_1> --B=<parent_2> <vcf>

Example

options:
  -h --help                   Show this screen.
  --version                   Show version.
  --vcf-out                   Output VCF instead of intervals.
  --endfill                   Don't leave gaps at the ends of chromosomes.
  --infill                    Fill in missing portions.
  --state=<state>             State probability [default: 0.97].
  --transition=<transition>   Transition probability [default: 1e-9]


"""
import sys
import itertools
from collections import defaultdict
from itertools import groupby
from operator import itemgetter
from signal import SIG_DFL, SIGPIPE, signal

import pomegranate
import matplotlib
import numpy as np

from clint.textui import colored, indent, puts_err
from docopt import docopt
from intervaltree import IntervalTree
from vcfkit.utils import autoconvert

from vcfkit.utils.fasta import *
from vcfkit.utils.vcf import *
from vcfkit.vk import __version__

matplotlib.use("Agg")

signal(SIGPIPE, SIG_DFL)


def generate_model(state, transition):
    # Setup hmm
    model = pomegranate.HiddenMarkovModel()

    A = pomegranate.State(pomegranate.DiscreteDistribution({'A': state, 'B': 1-state}), name='A')
    B = pomegranate.State(pomegranate.DiscreteDistribution({'A': 1-state, 'B': state}), name='B')

    model.add_transition(model.start, A, 0.5)
    model.add_transition(model.start, B, 0.5)

    model.add_transition(A, A, 1-transition)
    model.add_transition(A, B, transition)
    model.add_transition(B, A, transition)
    model.add_transition(B, B, 1-transition)

    model.add_transition(A, model.end, 0.5)
    model.add_transition(B, model.end, 0.5)

    model.bake(verbose=False)
    return model

to_model = {0: 'A', 1: 'B'}
to_gt = {0: '0/0', 1: '1/1'}
from_model = {True: 1, False: 0}

debug = None
if len(sys.argv) == 1:
    debug = ["hmm","--A=REF", "--B=ALT", "--vcf-out", "../test.vcf.gz"]


def generate_RLE(arr):
    grouped = [(k, sum(1 for i in g)) for k, g in groupby(arr)]
    return "".join([{0: "A", 1: "B"}[x] + str(y) for x, y in grouped]), len(grouped) - 1


if __name__ == '__main__':
    args = docopt(__doc__,
                  argv=debug,
                  options_first=False,
                  version=__version__)
    v = vcf(args["<vcf>"])
    # Put genotypes into large array
    """
        0/0 -> 0
        0/1 -> 1
        ./. -> 2
        1/1 -> 3
    """

    gt_list, dp_list, dv_list = [], [], []
    chromosome = []
    positions = []
    p1_gt_set = []
    p2_gt_set = []

    append_gt = True
    for n, line in enumerate(v):
        ref_alt = ["REF", "ALT"]
        if args["--A"] == "REF":
            p1_gt = 0
        else:
            p1_gt = line.gt_types[v.samples.index(args["--A"])]
        if args["--B"] == "ALT":
            p2_gt = 3
        else:
            p2_gt = line.gt_types[v.samples.index(args["--B"])]

        if p1_gt != p2_gt and p1_gt in [0,3] and p2_gt in [0,3]:
            chromosome.append(line.CHROM)
            positions.append(line.POS)
            gt_list.append(np.array(line.gt_types))
            dp_list.append(np.array(line.format("DP", int)))
            p1_gt_set.append(p1_gt)
            p2_gt_set.append(p2_gt)
    gt_set = np.vstack([gt_list]).astype("int8").T

    dp_set = np.vstack([dp_list]).astype("int32").T
    gt_set[gt_set == 1] = -9  # Het
    gt_set[gt_set == 2] = -9  # Missing
    
    gt_set[gt_set == 3] = 1  # Alt GT
    p1_gt_set = np.array(p1_gt_set)
    p2_gt_set = np.array(p2_gt_set)
    p1_gt_set[p1_gt_set == 3] = 1
    p2_gt_set[p2_gt_set == 3] = 1

    dp_set[dp_set < 0] = 0

    # Result dict
    vcf_gt_out = defaultdict(lambda: defaultdict(
        lambda: defaultdict(lambda: defaultdict(list))))
    sample_gt_matrix = np.array([])

    chrom_to_factor = {y: x for x, y in enumerate(set(chromosome))}
    sample_to_factor = {y: x for x, y in enumerate(v.samples)}
    hmm_gt_set = []

    if args["--vcf-out"] and args["<vcf>"] == "-":
        with indent(2):
            exit(puts_err(colored.blue("Cannot use vcf-out with stdin.")))

    if not args["--vcf-out"]:
        print("chrom\tstart\tend\tsample\tgt\tgt_name\tsupporting_sites\tsites\tDP\tswitches\trle")

    s = 0
    tree = {}
    for sample, column in zip(v.samples, gt_set):
        sample_gt = list(zip(chromosome, positions, column == p2_gt_set, column))
        sample_gt = [x[:3] for x in sample_gt if x[3] in [0, 1]]
        sequence = [to_model[x[2]] for x in sample_gt]

        model = generate_model(float(args['--state']), float(args['--transition']))
        if sequence:
            results = model.forward_backward(sequence)[1]
            if model.states[0].name == 'A':
                result_gt = [from_model[x]
                             for x in np.greater(results[:, 1], results[:, 0])]
            else:
                result_gt = [from_model[x]
                             for x in np.greater(results[:, 0], results[:, 1])]
            dp_s = dp_set[0][s]
            dp_s = dp_s[dp_s > 0]
            results = ((a, b, c, d, e)
                       for (a, b, c), d, e in zip(sample_gt, result_gt, dp_s))

            results = list(results)

            dp_avg = ""
            chrom_sets = [list(g) for k, g in groupby(results, itemgetter(0))]
            tree[sample] = {}
            for chrom_set in chrom_sets:
                end = 0
                chrom = chrom_set[0][0]
                tree[sample][chrom] = IntervalTree()
                interval_set = [list(g) for k, g in groupby(chrom_set, itemgetter(3))]
                interval_set_len = len(interval_set)
                for interval_n, interval in enumerate(interval_set):
                    orig = [x[2] for x in interval]
                    pred = [x[3] for x in interval]
                    gt = pred[0]
                    orig_RLE, switches = generate_RLE(orig)
                    supporting_sites = len([x for x in interval if x[2] == x[3]])
                    dp_avg = 0
                    if supporting_sites > 0:
                        dp_avg = sum([x[4] for x in interval if x[2] == x[3]])*1.0/supporting_sites
                    gt = interval[0][3]
                    if supporting_sites > 0:
                        if args["--infill"]:
                            start = end + 1
                        else:
                            start = min([x[1] for x in interval])
                        end = max([x[1] for x in interval])
                        sites = len(interval)

                        # End Fill
                        if args["--endfill"] and interval_n == 0:
                            start = 1
                        if args["--endfill"] and interval_n == interval_set_len - 1:
                            end = v.contigs[chrom]
                        gt_name = {0: args['--A'], 1: args['--B']}[gt]
                        output = [chrom,
                                  start,
                                  end,
                                  sample,
                                  gt + 1,
                                  gt_name,
                                  supporting_sites,
                                  sites,
                                  dp_avg,
                                  switches,
                                  orig_RLE]
                        if not args["--vcf-out"]:
                            out_line = "\t".join(map(str, output))
                            print(out_line)
                        else:
                            tree[sample][chrom][start:end+1] = gt
        s += 1

    if args["--vcf-out"]:
        v = vcf(args["<vcf>"])
        chrom_pos = [x[0:2] for x in sample_gt]
        # Add GT_ORIG and print raw header.
        v.add_format_to_header({"ID": "GT_ORIG",
                                "Description": "Original genotype",
                                "Type": "Character",
                                "Number": "1"})
        print((v.raw_header.strip()))
        for n, line in enumerate(v):
            line = variant_line(line, v.samples)
            for sample_n, sample in enumerate(v.samples):
                gt_orig = line.get_gt("GT", sample_n)
                try:
                    new_gt = next(iter(tree[sample][line.chrom].search(line.pos))).data
                except:
                    new_gt = None
                if new_gt is not None:
                    line.set_gt("GT_ORIG", sample_n, gt_orig)
                    line.set_gt("GT", sample_n, to_gt[new_gt])
            print(line)



#!/usr/bin/env python
"""
usage:
    vk primer template [options] <vcf>
    vk primer sanger   [options] <vcf>
    vk primer snip     [options] <vcf>
    vk primer indel    [options] <vcf>

options:
    -h --help                   Show this screen.
    --version                   Show version.
    --ref=<reference>           Reference Genome
    --region=<region>           Restrict to region.
    --samples=<samples>         Output genotypes for a sample or set of samples. [default: ALL]
    --template=<template>       The sample template to output [default: ALT]
    --size=<int>                Amplicon size [default: 600-800]
    --box-variants              Add second column for the sequence with the variant boxed.
    --polymorphic               Only output variants that are polymorphic across specified samples.
    --enzymes=<enzymes>         snip-SNP only: Specify groups of restriction enzymes or individual enzymes [default: Common]
    --nprimers=<nprimers>       Maximum number of primers to generate [default: 5]

"""
import sys
from docopt import docopt
from logzero import logger
from vcfkit.utils import message
from vcfkit.utils.primer_vcf import primer_vcf
from vcfkit.utils.reference import *
from vcfkit.utils.fasta import *
from vcfkit.utils import check_program_exists

from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE, SIG_DFL)

if len(sys.argv) == 1:
    debug = ['primer', "--ref=WBcel235", "test.vcf.gz"]

def main(debug=None):
    arguments = docopt(__doc__, 
                       argv=debug,
                       options_first=False)
    check_program_exists("primer3_core")
    check_program_exists("blastn")

    # Ensure user has specified a reference.
    if arguments["--ref"] is None:
        exit(message("Must specify a reference with --ref", color="red"))

    v = primer_vcf(arguments["<vcf>"],
                   reference=arguments["--ref"],
                   use_template=arguments['--template'],
                   polymorphic=arguments["--polymorphic"])
    v.enzymes = arguments['--enzymes']
    v.nprimers = int(arguments['--nprimers'])
    
    # Region
    if arguments["--region"]:
        v.region = arguments["--region"]
    else:
        v.region = None

    # Samples
    if arguments["--samples"]:
        if arguments["--samples"] == "ALL":
            v.output_samples = v.samples
        else:
            v.output_samples = arguments["--samples"].split(",")
            for sample in v.output_samples:
                if sample not in v.samples + ["REF", "ALT"]:
                    exit(message(f"{sample} not found in VCF", color = "red"))

    v.box_variants = arguments["--box-variants"] # Needs to be implemented.

    # Check for std. input
    if arguments["template"]:
        v.mode = "template"
        if '-' in arguments['--size']:
            try:
                v.region_size = int(arguments['--size'].split("-")[0])
                message("Warning: region size set to {s}".format(s=v.region_size))
            except:
                exit(message("Error: Invalid --size"))
        elif str.isdigit(arguments['--size']):
            v.region_size = int(arguments['--size'])
        v.amplicon_lower = 0
        v.amplicon_upper = 0
    elif arguments["indel"]:
        message("--size ignored; size is set dynamically when genotyping indels.")
        v.mode = "indel"
    elif arguments["snip"]:
        message("--size ignored; Set to 600-800 bp.")
        v.mode = "snip"
        v.amplicon_lower = 600
        v.amplicon_upper = 800
        v.region_size = 500 #x2=1000

    elif arguments["sanger"]:
        v.mode = "sanger"
        if arguments['--size'] is None:
            size = "600-800"
            message("Warning: --size set to 600-800 for sanger sequencing.")
        elif str.isdigit(arguments['--size']):
            exit(message("Error: You must specify a amplicon --size range for Sanger; e.g. 600-800."))
        else:
            size = arguments['--size']
        
        if "-" in size:
            try:
                v.amplicon_lower = int(size.split("-")[0])
                v.amplicon_upper = int(size.split("-")[1])
            except:
                exit(message("Error: Invalid --size"))
        else:
            v.amplicon_lower = 600
            v.amplicon_upper = 800
        v.region_size = (v.amplicon_upper//2) + 100
        if (v.amplicon_lower < 300 or 800 < v.amplicon_upper):
            message("Warning: --size range should (probably) be between 300-800 for sanger sequencing.")

    for variant in v.variant_iterator():
        variant.out()

if __name__ == '__main__':
    main()




import os
import re
import sys
from cyvcf2 import VCF as cyvcf2
from collections import OrderedDict, deque
from itertools import islice
from vcfkit.utils import message
from copy import copy
import numpy as np
from .reference import resolve_reference_genome
np.set_printoptions(threshold=sys.maxsize)


class vcf(cyvcf2):
    def __init__(self, filename, reference=None):
        if not os.path.isfile(filename) and filename != "-":
            exit(message("Error: " + filename + " does not exist"))
        self.filename = filename
        if reference:
            self.reference = reference
            self.reference_file = resolve_reference_genome(reference)

        cyvcf2.__init__(self, self.filename)
        # Check if file exists
        self.n = len(self.samples)  # Number of Samples

        # Meta Data
        comp = re.compile(r'''^##(?P<key>[^<#]+?)=(?P<val>[^<#]+$)''', re.M)
        self.metadata = OrderedDict(comp.findall(self.raw_header))

        # Contigs
        self.contigs = OrderedDict(list(zip(
            re.compile("##contig=<ID=(.*?),").findall(self.raw_header),
            list(map(int, re.compile("##contig.*length=([^,>]*?)>").findall(self.raw_header)))
        )))

        self.info_set = [x for x in self.header_iter() if x.type == "INFO"]
        self.filter_set = [x for x in self.header_iter() if x.type == "FILTER"]
        self.format_set = [x for x in self.header_iter() if x.type == "FORMAT"]
        self.header = copy(self.raw_header)

    def window(self, shift_method, window_size, step_size=None):
        """
            Generates windows of VCF data by positions or SNP
        """
        result_list = variant_interval(window_size=window_size, step_size=step_size, shift_method=shift_method)
        try:
            while True:
                line = next(self)
                #
                # SNP-Sliding and SNP-Interval
                #
                if shift_method in ["SNP-Sliding", "SNP-Interval"]:
                    result_list.append(line)
                    if not result_list.unique_chroms():
                        yield result_list[0:len(result_list)-1]
                        result_list.get_last()
                    elif len(result_list) == window_size:
                        yield result_list
                        if shift_method == "SNP-Interval":
                            result_list.clear()
                #==============#
                # POS-Interval #
                #==============#
                elif shift_method == "POS-Interval":
                    result_list.append(line)
                    last_chrom = line.CHROM
                    chrom = line.CHROM
                    while True:
                        if chrom != last_chrom:
                            yield result_list.filter_within_bounds()
                            result_list.get_last()
                            result_list.lower_bound = 0
                            result_list.upper_bound = window_size
                            chrom = line.CHROM
                        if line.POS < result_list.lower_bound:
                            result_list.iterate_interval()
                        elif line.POS >= result_list.upper_bound:
                            yield result_list.filter_within_bounds()
                            result_list.iterate_interval()
                        if result_list.lower_bound <= line.POS < result_list.upper_bound:
                            line = next(self)
                            result_list.append(line)
                        last_chrom = line.CHROM
                #=============#
                # POS-Sliding #
                #=============#
                elif shift_method == "POS-Sliding":
                    result_list.append(line)
                    result_list.lower_bound = result_list[0].POS - window_size/2
                    if result_list.lower_bound < 0:
                        result_list.lower_bound = 0
                    result_list.upper_bound = result_list.lower_bound + window_size
                    while True:
                        positions = result_list.positions()
                        max_pos = max(positions)
                        min_pos = min(positions)
                        chrom_num = len(set([o.CHROM for o in result_list]))
                        if (max_pos - min_pos) > window_size or chrom_num > 1:
                            result_list.popleft()
                            result_list.iterate_interval()  # Updates lower and upper bounds
                        else:
                            break
                    # Update variant interval
                    yield result_list

        except StopIteration:
            if len(result_list) > 0:
                if shift_method == "POS-Interval":
                    yield result_list.filter_within_bounds()
                else:
                    yield result_list

    def output_raw(self):
        """
            Outputs raw vcf
        """
        for hline in self.raw_header.strip("\n").split("\n"):
            yield hline
        for line in self:
            yield(str(line))


class variant_interval(deque):
    def __init__(self, interval=[], window_size=None, step_size=None, shift_method=None, varlist=None, lower_bound=0):
        self.shift_method = shift_method
        self.window_size = window_size
        self.step_size = step_size
        if self.step_size is None:
            self.step_size = window_size
        self.lower_bound = lower_bound
        self.upper_bound = self.lower_bound + window_size
        if shift_method in ["SNP-Sliding", "SNP-Interval"]:
            super(variant_interval, self).__init__(interval, maxlen=self.window_size)
        else:
            super(variant_interval, self).__init__(interval)

    def positions(self):
        return [x.POS for x in self]

    def iterate_interval(self):
        if self.shift_method == "POS-Sliding":
            mean_pos = np.mean(self.positions())
            half_window = self.window_size/2
            self.lower_bound = int(mean_pos - half_window)
            if self.lower_bound < 0:
                self.lower_bound = 0
            self.upper_bound = int(mean_pos + half_window)
        else:
            self.lower_bound += self.step_size
            self.upper_bound += self.step_size
        self.CHROM = self[0].CHROM
        return self

    def filter_within_bounds(self):
        cut = [x for x in self if x.POS >= self.lower_bound and x.POS < self.upper_bound]
        return variant_interval(interval=cut,
                                window_size=self.window_size,
                                shift_method=self.shift_method,
                                lower_bound=self.lower_bound,
                                step_size=self.step_size)

    def __getitem__(self, index):
        if isinstance(index, slice):
            cut = islice(self, index.start, index.stop, index.step)
            return variant_interval(interval=cut,
                                    window_size=self.window_size,
                                    shift_method=self.shift_method,
                                    lower_bound=self.lower_bound,
                                    step_size=self.step_size)
        return deque.__getitem__(self, index)

    def get_last(self):
        """
            Removes all items from the variant interval and
            sets to last item
        """
        super(variant_interval, self).__init__(self[len(self)-1:len(self)], maxlen=self.window_size)

    def unique_chroms(self):
        """
            Checks to see whether the variants
            all belong to the same chromosome
        """
        return len(set([var.CHROM for var in self])) == 1 and len(self) > 0


class variant_set:
    def __init__(self, ins, samples):
        if type(ins) == str:
            if samples:
                self.variants = [variant_line(x, samples) for x in ins.splitlines()]
            else:
                self.variants = [variant_line(x) for x in ins.splitlines()]


class variant_line:
    """
        Used to modify genotypes by HMM currently...
    """
    def __init__(self, line, sample_names=None):
        self.line = str(line).strip().split("\t")
        self.chrom = self.line[0]
        self.pos = int(self.line[1])
        self.chrom_pos = (self.chrom, self.pos)
        self.format_keys = self.line[8].split(":")
        self.values = [x.split(":") for x in self.line[9:]]

    def get_gt(self, format_field, index):
        f_index = self.format_keys.index(format_field)
        return self.values[index][f_index]

    def set_gt(self, format_field, index, value):
        if format_field not in self.format_keys:
            self.format_keys += [format_field]
            self.values = [x + ["."] for x in self.values]
        f_index = self.format_keys.index(format_field)
        self.values[index][f_index] = value

    def __str__(self):
        self.line = self.line[0:8] + [":".join(self.format_keys)] + [":".join(x) for x in self.values]
        return '\t'.join(self.line)



#! /usr/bin/env python
"""
usage:
  tb annotate <vcf> 

"""
import re
from docopt import docopt
from subprocess import Popen, PIPE
from signal import signal, SIGPIPE, SIG_DFL

from vcfkit.utils import autoconvert
from vcfkit.utils.vcf import *
from vcfkit.utils.matrix import *

signal(SIGPIPE, SIG_DFL)

debug = None



ann_fields = ["allele", 
              "annotation",
              "putative_impact",
              "gene_name",
              "gene_id",
              "feature_type",
              "feature_id",
              "transcript_biotype",
              "rank_total",
              "hgvs_c",
              "hgvs_p",
              "cdna_position",
              "cds_position",
              "protein_position",
              "distance_to_feature",
              "errors"]

def parse_aa(hgvs_p):
    aa = re.split("\\.|[0-9]+", hgvs_p)[1:]
    aa = [three_to_one_aa[x] for x in aa]
    return tuple(aa)


def calc_grantham(aa_set):
    """
        Calculate Grantham Score
    """
    ret_set = []
    for i in aa_set:
        if i.count("*") == 0:
            ret_set.append(grantham[i])
        else:
            ret_set.append("")
    return "|".join(ret_set)


if __name__ == '__main__':
    args = docopt(__doc__,
                  version='VCF-Toolbox v0.1',
                  argv=debug)
    if args["<vcf>"] == "":
        print(__doc__)
    v = vcf(args["<vcf>"])
    v.add_info_to_header({"ID": "grantham", "Description": "Grantham Score", "Type": "String", "Number": 1})
    for line in v:
        if "ANN" in dict(line.INFO):
            ANN = dict(line.INFO)["ANN"]
            ANN = [dict(list(zip(ann_fields, x.split("|")))) for x in ANN.split(",")]
            aa_set = [parse_aa(x["hgvs_p"]) for x in ANN]
            line.INFO["grantham"] = calc_grantham(aa_set)
            print(line)





#!/usr/bin/env python
"""
usage:
    vk call <seq> --ref=<reference> (--all-sites|--vcf-sites) <vcf>

options:
    -h --help                   Show this screen.
    --version                   Show version.

"""
import os
import sys
from collections import defaultdict
from signal import SIG_DFL, SIGPIPE, signal

from Bio import SeqIO
from Bio.Blast.Applications import NcbiblastxCommandline
from clint.textui import colored, indent, puts, puts_err
from docopt import docopt
from vcfkit.utils.blastn import blast, blast_variant
from vcfkit.utils.reference import resolve_reference_genome
from vcfkit.utils.vcf import *

from vcfkit import vk

signal(SIGPIPE, SIG_DFL)


def seq_type(filename):
    """
        Resolves sequence filetype using extension.
    """
    filename, ext = os.path.splitext(filename.lower())
    if ext in [".fasta", ".fa"]:
        extension = 'fasta'
    elif ext in [".fastq",".fq"]:
        extension = 'fastq'
    elif ext in [".ab1", '.abi']:
        extension = 'abi'
    else:
        raise Exception("Unknown sequence file type: " + filename)

    with indent(4):
        puts_err(colored.green("\nReading sequences as %s\n" % extension.upper()))
    return extension


def resolve_sample_from_line(samples, line):
    """
        Resolves sample names by splitting fasta line
        on non-word characters.
    """
    line = re.split("[ \t]", line)
    matched_sample = [x for x in samples if x in line]
    if len(matched_sample) == 1:
        return matched_sample[0]
    return ""


def format_gt(gt):
    # Return homozygous calls as single bases.
    gt = list(set(re.split("[|/]", gt)))
    if len(gt) == 1:
        return gt[0]
    else:
        return '/'.join(gt)

def format_args(args, add_missing_stdin = False):
    if add_missing_stdin:
        pass # Test for vcf

def main(debug=None):
    args = docopt(__doc__,
                  version='VCF-Toolbox v0.1',
                  argv=debug,
                  options_first=False)


    module_path = os.path.split(os.path.realpath(__file__))[0]
    handle = open(args["<seq>"], "rb")
    reference = resolve_reference_genome(args["--ref"])

    if args["<vcf>"]:
        concordance = True
        v = vcf(args["<vcf>"])
        samples = v.samples


    if args["--vcf-sites"] and args["<vcf>"] is None:
        with indent(4):
            exit(puts_err(colored.red("\nMust specify <vcf> with --vcf-sites\n")))

    # Setup reference for blast call
    b = blast(reference)

    # Set file type:
    sequence_file_type = seq_type(args["<seq>"])

    # Output header
    print(("\t".join(blast_variant.output_order)))
    for record in SeqIO.parse(handle, sequence_file_type):
        # Resolve sample within fasta line
        sample = resolve_sample_from_line(samples, handle.name)
        if not sample:
            sample = resolve_sample_from_line(samples, record.name)
        blast_results = b.blast_call(record)
        classification = ""
        for n, variant in enumerate(blast_results):
            output_line = False
            if variant is None:
                puts_err(colored.red("No Results for " + sample + " " + record.description))
                continue
            if args["<vcf>"]:
                if n == 0:
                    vcf_variants = []
                    for vcf_variant in v(variant.region()):
                        if sample:
                            gt = format_gt(vcf_variant.gt_bases[v.samples.index(sample)])
                            vcf_variants.append([vcf_variant.CHROM,
                                                 vcf_variant.POS,
                                                 gt,
                                                 vcf_variant.REF,
                                                 vcf_variant.ALT])
                            vcf_variant_positions = [x[0:2] for x in vcf_variants]

                chrom_pos =  variant.chrom_pos_allele()[0:2]
                vcf_variant_match = [x for x in vcf_variants if x[0:2] == chrom_pos]
                if vcf_variant_match:
                    vcf_variant_match = vcf_variant_match[0]
                    variant.vcf_gt = vcf_variant_match[2]
                    variant.REF = vcf_variant_match[3]
                    variant.ALT = ','.join(vcf_variant_match[4])
                    variant.fetch_variant_type()
                    if variant.REF == variant.seq_gt and variant.seq_gt == variant.vcf_gt:
                        variant.classification = "TN"
                    elif variant.REF != variant.seq_gt and variant.seq_gt == variant.vcf_gt:
                        variant.classification = "TP"
                    elif variant.REF == variant.seq_gt and variant.seq_gt != variant.vcf_gt:
                        variant.classification = "FP" 
                    elif variant.REF != variant.seq_gt and variant.seq_gt != variant.vcf_gt:
                        variant.classification = "FN"
                else:
                    variant.REF = ""
                    variant.ALT = ""
                    variant.fetch_variant_type()
                    variant.classification = ""

                if args["--vcf-sites"] and variant.classification != "":
                    output_line = True
                elif args["--all-sites"] is True:
                    output_line = True
            else:
                if args["--all-sites"]:
                    output_line = True
                elif variant.is_variant:
                    output_line = True
            if output_line:

                variant.sample = sample
                if record.description:
                    variant.description = record.description
                else:
                    variant.description = os.path.split(handle.name)[1]
                print('\t'.join([str(variant)]))

if __name__ == '__main__':
    main()



#! /usr/bin/env python
"""
usage:
  vk rename [--prefix=<prefix> --suffix=<suffix> --subst=<subst>...] <vcf>

options:
  -h --help                   Show this screen.
  --version                   Show version.

"""
import re
from signal import SIG_DFL, SIGPIPE, signal

from docopt import docopt
from vcfkit import __version__

from vcfkit.utils.vcf import *

signal(SIGPIPE, SIG_DFL)


def main(debug=None):
    args = docopt(__doc__,
                  argv=debug,
                  version=__version__)
    if args["<vcf>"] == "":
        print(__doc__)
    v = vcf(args["<vcf>"])
    for line in v.output_raw():
        if line.startswith("#CHROM"):
            line = line.split("\t")
            if args["--subst"]:
                find_replace = [re.split("[:=,]", x) for x in args["--subst"]]
                for orig, replacement in find_replace:
                    for n, sample in enumerate(line[9:]):
                        if sample == orig:
                            line[9+n] = replacement
            if args["--prefix"]:
                line[9:] = [args["--prefix"] + x for x in line[9:]]
            if args["--suffix"]:
                line[9:] = [x + args["--suffix"] for x in line[9:]]
            print('\t'.join(line))
        else:
            print((line.strip()))

if __name__ == '__main__':
    main()



#! /usr/bin/env python
"""
usage:
  vk <command> [<args>...] 
  vk setup
  vk -h | --help
  vk --version

commands:
  calc
  call
  filter
  geno
  genome
  hmm
  phylo
  primer
  rename
  tajima
  vcf2tsv

"""
from vcfkit import __version__
from .utils import lev, message
from docopt import docopt
from subprocess import call, check_output, CalledProcessError
from .utils.vcf import *
from clint.textui import colored, puts, indent
import sys
from . import vk
import os
import signal
signal.signal(signal.SIGINT, lambda x,y: sys.exit(0))
command_list = [x.strip() for x in filter(len, __doc__.splitlines()[8:])]



debug = None
if len(sys.argv) == 1:
    debug = [""]


def getScriptPath():
    return os.path.dirname(vk.__file__)


def main():
    args = docopt(__doc__,
                  argv=debug,
                  options_first=True,
                  version=__version__)
    argv = [args['<command>']] + args['<args>']
    program_list = {"bwa": "bwa",
                    "samtools": "samtools",
                    "bcftools": "bcftools",
                    "blast": "blastn",
                    "muscle": "muscle"}
    if args["<command>"] == "setup":
        """
            Use Homebrew to install programs!
        """
        program_installed = list(program_list.keys())
        for install_name, program in list(program_list.items()):
            try:
                with indent(4):
                    puts(colored.blue("Installing " + install_name))
                check_output(["brew", "install", install_name])
                program_installed.remove(install_name)
            except CalledProcessError:
                try:
                    check_output(["which", program])
                    with indent(4):
                        puts(colored.blue(program + " previously installed"))
                    program_installed.remove(install_name)
                except CalledProcessError:
                    with indent(4):
                        puts(colored.red("Error installing " + install_name))
        if len(program_installed) == 0:
            with indent(4):
                puts(colored.blue("Programs successfully installed!"))
        else:
            with indent(4):
                puts(colored.red("Error: Not all programs successfully installed: " + ", ".join(program_installed)))
    elif args["<command>"] == "":
        print(__doc__)
        for prog in list(program_list.values()):
            try:
                check_output(["which", prog])
            except CalledProcessError:
                with indent(4):
                    puts(
                        colored.red(prog + " not installed. Use a package manager to install or try using 'vk setup'\n"))
    elif args['<command>'] in command_list:
        comm = ['python', getScriptPath() + '/' + args["<command>"] + ".py"] + argv
        exit(call(comm))
    else:
        levs = [(x, lev(args['<command>'], x)) for x in command_list]
        closest =  min(levs, key = lambda x: x[1])[0]
        command = args['<command>']
        message("There is no command '{command}'. Did you mean 'vk {closest}'?".format(**locals()))

if __name__ == '__main__':
    main()



from subprocess import Popen, PIPE
from vcfkit.utils import *
from collections import defaultdict, OrderedDict
from pprint import pprint as pp
import csv
from clint.textui import puts_err
from Bio.Blast import NCBIXML
from Bio.SeqRecord import SeqRecord
from copy import copy
from io import StringIO
from vcfkit.utils import *
import sys
import signal
signal.signal(signal.SIGINT, lambda x,y: sys.exit(0))



def clamp(n, minn, maxn):
    return max(min(maxn, n), minn)

def fastq_mean(x):
    if len(x) > 0:
        return round(sum(x)*1.0/len(x), 2)
    else:
        return 0


class blast_variant:
    """
        Class representing a blast variant
    """
    output_order = ["CHROM",
                    "POS",
                    "reference",
                    "REF",
                    "ALT",
                    "seq_gt",
                    "vcf_gt",
                    "sample",
                    "variant_type",
                    "classification",
                    "index",
                    "alignment_start",
                    "alignment_end",
                    "strand",
                    "context",
                    "gaps",
                    "mismatch",
                    "evalue",
                    "bitscore",
                    "phred_quality",
                    "phred_quality_window",
                    "description"]

    def __init__(self,
                 blast_result,
                 POS,
                 reference_seq,
                 query_seq,
                 seq_out,
                 index,
                 reference_index,
                 context,
                 gaps,
                 mismatch,
                 phred_quality = "",
                 phred_quality_window = ""):

        self.CHROM = blast_result["sacc"]
        self.POS = POS
        posp1 = self.POS + 1
        self.CHROM_POS = "{self.CHROM}:{self.POS}-{posp1}".format(**locals())
        self.classification = None
        self.reference = reference_seq[reference_index]
        self.seq_gt = seq_out.strip("-")
        self.vcf_gt = ""
        self.alignment_start = blast_result["sstart"]
        self.alignment_end = blast_result["send"]
        self.strand = {"plus": "+", "minus": "-"}[blast_result["sstrand"]]
        self.context = context
        self.gaps = gaps
        self.mismatch = mismatch
        self.index = index
        self.evalue = blast_result["evalue"]
        self.bitscore = blast_result["bitscore"]
        self.phred_quality = phred_quality
        self.phred_quality_window = phred_quality_window

    def fetch_variant_type(self):
        if self.REF != self.ALT:
            if len(self.REF) == len(self.seq_gt):
                variant_type = "snp"
            elif len(self.REF) > len(self.seq_gt):
                variant_type = "deletion"
            elif len(self.REF) < len(self.seq_gt):
                variant_type = "insertion"
        else:
            variant_type = ""
        self.variant_type = variant_type

    def chrom_pos_allele(self):
        """
            Simplified representation of variant
        """
        return [self.CHROM, self.POS, self.seq_gt]

    def region(self):
        return "{self.CHROM}:{self.alignment_start}-{self.alignment_end}".format(**locals())

    def __str__(self):
        return '\t'.join([str(getattr(self, x)) for x in blast_variant.output_order])


class blast:
    def __init__(self, db, num_alignments=1, word_size = 28):
        self.db = db
        self.num_alignments = num_alignments
        self.word_size = word_size
        self.output_format = ["sacc",
                              "pident",
                              "gaps",
                              "mismatch",
                              "length",
                              "qstart",
                              "qend",
                              "sstart",
                              "send",
                              "evalue",
                              "bitscore",
                              "qseq",
                              "sseq",
                              "sstrand"]
        self.output_string = ' '.join(self.output_format)
        self.blastn_query_str = ' '.join(["echo {self.query} | blastn -query - -db={self.db} ",
                             "-outfmt '6 {self.output_string}'",
                             "-num_alignments {self.num_alignments}",
                             "-word_size {self.word_size}"])

    def fetch_reference_seq(self, chrom, start, end):
        end += 20
        fetch_ref_cmd = ["samtools",
         "faidx",
         self.db,
         "{chrom}:{start}-{end}".format(**locals())]
        resp, err = Popen(fetch_ref_cmd,
               stdout=PIPE,
               stderr=PIPE).communicate()
        return ''.join(resp.decode('utf-8').splitlines()[1:]).upper()


    def print_alignment(self, q, start=0, end=None):
        """ 
            Print the alignment with matches
        """
        blast_result = self.blast_search(q)
        if end is None:
            end = len(blast_result["sseq"])
        sseq = blast_result["sseq"]
        qseq = blast_result["qseq"]
        spacer = []
        for i in range(len(sseq)):
            if qseq[i] == sseq[i]:
                spacer.append("|")
            else:
                spacer.append(" ")
        spacer = ''.join(spacer)
        print((blast_result["sseq"][start:end]))
        print((spacer[start:end]))
        print((blast_result["qseq"][start:end]))

    def blast_search(self, q):
        """
            Perform a blast search and store the result
        """
        if type(q) == SeqRecord:
            self.query = q.seq
            if hasattr(q, "letter_annotations"):
                if 'phred_quality' in list(q.letter_annotations.keys()):
                    self.query_qual = q.letter_annotations["phred_quality"]
                else:
                    self.query_qual = None
        else:
            self.query = q

        self.query_length = len(q)
        blastn_query = self.blastn_query_str.format(**locals())
        resp, err = Popen(blastn_query,
                          stdout=PIPE,
                          stderr=PIPE,
                          shell=True).communicate()
        if not resp:
           return None
        if err:
           raise Exception(err)

        # Format variables
        resp = [OrderedDict(list(zip(self.output_format,
                                list(map(autoconvert, x.split("\t")))))) 
                for x in resp.decode("utf-8").splitlines()]
        return resp

    def check_primer(self, q):
        blast_results = self.blast_search(q)
        if blast_results:
            return len([i for i in blast_results if i['length'] > 14])
        else:
            return 0


    def blast_call(self, q):
        """
            Call variants from blast alignments
        """
        blast_result = self.blast_search(q)
        if blast_result is None:
            yield None
        else:
            for bresult in blast_result:
                CHROM = bresult["sacc"]
                start = bresult["sstart"]
                end = bresult["send"]
                ref = bresult["sseq"]
                alt = bresult["qseq"]
                gaps = bresult["gaps"]
                mismatch = bresult["mismatch"]

                # Retrieve reference sequence
                reference_seq = self.fetch_reference_seq(CHROM, start, end)

                # Compare fasta sites
                ref_out = ""
                seq_out = ""
                len_insertions = 0
                len_deletions = 0
                phred_quality = ""
                phred_quality_window = ""
                i = 0
                while i < len(ref)-1:
                    if alt[i+1] == "-":
                        ref_out = ref[i]
                        seq_out = alt[i]
                        while alt[i+1] == "-":
                            i += 1
                            ref_out += ref[i]
                            seq_out += alt[i]
                        len_deletions += len(seq_out) - 1
                    elif ref[i+1] == "-":
                        ref_out = ref[i]
                        seq_out = alt[i]
                        while ref[i+1] == "-":
                            i += 1
                            ref_out += ref[i]
                            seq_out += alt[i]
                        len_insertions += len(ref_out) - 1
                    else:
                        ref_out, seq_out = ref[i], alt[i]
                    

                    POS = i - len_insertions
                    context_start = clamp(i-10,0,len(ref))
                    context_end = clamp(i+10,0,len(ref))
                    context = alt[context_start:i] + "[" + seq_out + "]" + alt[i+1:context_end]
                    context = context.replace("-","")
                    index = i + bresult["qstart"] - len_deletions - len(seq_out)

                    if 'qqual' in bresult:
                        window_start = clamp(i-6, 0, len(bresult["qqual"]))
                        window_end = clamp(i+7, 0, len(bresult["qqual"]))
                        phred_quality_window = fastq_mean([x for x in bresult["qqual"][window_start:window_end] if x != "NA"])
                        phred_quality = bresult["qqual"][i]

                    variant_out = blast_variant(blast_result=bresult,
                                                POS=POS + start,
                                                reference_index=POS,
                                                index=index,
                                                seq_out=seq_out,
                                                query_seq=q,
                                                reference_seq=reference_seq,
                                                context=context,
                                                gaps=gaps,
                                                mismatch=mismatch,
                                                phred_quality=phred_quality,
                                                phred_quality_window=phred_quality_window)
                    yield variant_out
                    i += 1




#! /usr/bin/env python
"""
usage:
  vk vcf2sql (postgres|mysql|sqlite) [options] <vcf>

options:
  -h --help                   Show this screen.
  --version                   Show version.

  --db=<db>                   Database Name for MySQL, Postgres OR filename for sqlite
  --user=<user>               User for MySQL, Postgres
  --password=<pw>             Password for MySQL, Postgres  
  --host=<host>               Host for MySQL, Postgres
  --compress                  Compress GT data.

  --table-name=<table-name>   Append a prefix to table names
  --vcf-version               Create a column indicating VCF version
  --simple                    Use Reduced Field Set
  --ANN                       Parse snpeff fields
  --modifier                  Include modifier annotation records.
  --print                     Print CSV to stdout instead of loading.

"""
from docopt import docopt
from .utils.vcf import * 
import sys
import os
import re
from peewee import *
from signal import signal, SIGPIPE, SIG_DFL
from slugify import slugify
import datetime
import copy
import tempfile
import csv
import json
import zlib, pickle, base64
from playhouse.csv_loader import load_csv
signal(SIGPIPE, SIG_DFL)


def boolify(s):
    if s == 'True':
        return True
    if s == 'False':
        return False
    raise ValueError("huh?")


def autoconvert(s):
    for fn in (boolify, int, float):
        try:
            return fn(s)
        except ValueError:
            pass
    return s

load_bigquery = r"""
#!/usr/bin/bash
# Load Bigquery

# [ ] Add instructions on getting set up...

gsutil cp {tsv_out} gs://andersen/annotation/{tsv_out}

# Load on bigquery
bq load --field_delimiter "\t" andersen-lab:Variation.{vcf_safe} gs://andersen/{tsv_out} {bigquery_schema}

"""

# Standard Columns
standard = (("CHROM", 1, "String", "Chromosome/Contig", "STANDARD"),
            ("POS", 1, "Integer", "Chromosome/Contig Position", "STANDARD"),
            ("_ID", 1, "String", "Variant ID", "STANDARD"),
            ("REF", 1, "String", "Reference Allele", "STANDARD"),
            ("ALT", 1, "String", "Alternative Alleles (list)", "STANDARD"),
            ("QUAL", 1, "Float", "Variant Quality", "STANDARD"),
            ("FILTER", 1, "String", "Variant Filters", "STANDARD"))


# Info
r_info = re.compile(r'''\#\#INFO=<
  ID=(?P<id>[^,]+),
  Number=(?P<number>-?\d+|\.|[AG]),
  Type=(?P<type>Integer|Float|Flag|Character|String),
  Description="(?P<desc>[^"]*)".*
  >''', re.VERBOSE)

# Format
r_format = re.compile(r'''\#\#FORMAT=<
  ID=(?P<id>.+),
  Number=(?P<number>-?\d+|\.|[AG]),
  Type=(?P<type>.+),
  Description="(?P<desc>.*)".*
  >''', re.VERBOSE)


try:
    os.remove("test.db")
except:
    pass

index_fields = ["CHROM", "POS", "SAMPLE", "FT", "_ID", "REF", "ALT", "gene_name", "gene_id", "putative_impact", "feature_type"]
simple_fields = ["FT", "GT", "TGT", "INDEL", "ANN"]
format_field_conversion = {"Integer": int, "String": str, "Float": float}

ann_fields = ["allele", 
              "annotation",
              "putative_impact",
              "gene_name",
              "gene_id",
              "feature_type",
              "feature_id",
              "transcript_biotype",
              "rank_total",
              "hgvs_c",
              "hgvs_p",
              "cdna_position",
              "cds_position",
              "protein_position",
              "distance_to_feature",
              "errors"]


ann_field_types = [str]*14


def filter_format_null(val):
    if val == ".":
        return None
    else:
        return val


def remove_missing_fields(val):
    ret_fields = []
    for k, v in val:
        if k in [x[0] for x in format_cols]:
            ret_fields.append((k, v))
    return ret_fields


def format_tgt(val, gt_dict):
    if val["GT"] == "./.":
        return None
    tgt = '/'.join([gt_dict[int(x)] for x in re.split("[\|/]", val["GT"])])
    return tgt

debug = None
if len(sys.argv) == 1:
    debug = ['vcf2sql', "test.vcf.gz"]

if __name__ == '__main__':
    args = docopt(__doc__,
                  argv=debug,
                  options_first=False)

    timestamp = datetime.datetime.now()

    module_path = os.path.split(os.path.realpath(__file__))[0]
    v = vcf(args["<vcf>"])
    vcf_safe = v.filename.replace(".", "_")
    tsv_out = v.filename.replace("vcf", "tsv").replace(
        "bcf", "tsv").replace(".gz", "") + ".gz"

    info_cols = [list(map(autoconvert, list(x))) + ["INFO"]
                 for x in r_info.findall(v.raw_header)]
    format_cols = [list(map(autoconvert, list(x))) + ["FORMAT"]
                   for x in r_format.findall(v.raw_header)]

    if args["--simple"]:
        info_cols = [x for x in info_cols if x[0] in simple_fields]
        format_cols = [x for x in format_cols if x[0] in simple_fields]

    if args["sqlite"]:
        db = SqliteDatabase(args["--db"])
    elif args["mysql"]:
        db = MySQLDatabase(args["--db"],
                           user = args["--user"],
                           password = args["--password"],
                           host = args["--host"])

    if not args["--print"]:
        db.connect()
        # Set table name
        tbl_name = slugify(args["<vcf>"])
        if args["--table-name"]:
            tbl_name = args["--table-name"]


    class vcf_table(Model):
        Variant = IntegerField(index = True)
        CHROM = CharField(index = True)
        POS = IntegerField(index = True)
        _ID = CharField(index = True, null = True)
        REF = CharField(index = True)
        ALT = CharField(index = True, null = True)
        QUAL = FloatField(null = True)
        FILTER = CharField(null = True)
        GT = TextField(null = True)

        if args["--ANN"]:
            allele = CharField(index=True, null = True)
            annotation = CharField(index=True, null = True)
            putative_impact = CharField(null=True)
            gene_name = CharField(index=True, null=True)
            gene_id = CharField(index=True, null=True)
            feature_type = CharField(null=True)
            feature_id = CharField(null=True)
            transcript_biotype = CharField(null=True)
            rank_total = CharField(null=True)
            hgvs_c = CharField(null=True)
            hgvs_p = CharField(null=True)
            cdna_position = CharField(null=True)
            cds_position = CharField(null=True)
            protein_position = CharField(null=True)
            distance_to_feature = CharField(null=True)
            errors = CharField(null=True)
        if not args["--print"]:
            class Meta:
                database = db
                db_table = tbl_name

    # Add in INFO Cols
    for i in info_cols:
        index_field = i[0] in index_fields
        if i[0] == "ANN":
            break
        if i[1] > 1:
            i[2] = "String"
        if i[2] == "String":
            CharField(index=index_field, null=True,
                      help_text=i[3]).add_to_class(vcf_table, i[0])
        elif i[2] == "Integer":
            IntegerField(index=index_field, null=True,
                         help_text=i[3]).add_to_class(vcf_table, i[0])
        elif i[2] == "Float":
            FloatField(index=index_field, null=True,
                       help_text=i[3]).add_to_class(vcf_table, i[0])
        elif i[2] == "Flag":
            BooleanField(null = True).add_to_class(vcf_table, i[0])

    # Output Schema
    with open("schema.sql", "w") as f:
        f.write("".join(vcf_table.sqlall()))

    db.create_tables([vcf_table], safe = True)
    c = 0
    loc_set = [] # 
    insert_set = [] # 
    csv_writer = csv.writer(sys.stdout, quoting=csv.QUOTE_MINIMAL)
    for variant_index, loc in enumerate(v):
        c += 1
        site_fields = {}
        annotation_record_set = []

        # Propogate standard fields
        for x in standard:
            if x[0] == "_ID":
                attr_name = "ID"
            else:
                attr_name = x[0]
            attr = getattr(loc, attr_name)
            if type(attr) == list:
                attr = ','.join(attr)
            site_fields[x[0]] = attr

        annotation_record_set = []
        # Propogate INFO fields
        for x in info_cols:
            # Add Annotation Fields
            if x[0] == "ANN":
                annotation_record = {}
                for ann in loc.INFO["ANN"].split(","):
                    for field_name, field_type, field_value in zip(ann_fields, ann_field_types, ann.split("|")):
                        if not field_value:
                            annotation_record[field_name] = None
                        else:
                            annotation_record[field_name] = field_type(field_value)
                    if ((annotation_record["putative_impact"] != "MODIFIER") or args["--modifier"]):
                        annotation_record_set.append(copy.copy(annotation_record))
            else:
                try:
                    attr = loc.INFO[x[0]]
                    if type(attr) == tuple or type(attr) == list:
                        attr = ','.join(map(str, attr))
                    site_fields[x[0]] = attr
                except:
                    site_fields[x[0]] = None

        # Insert genotypes
        gt_dict = {0: loc.REF}
        alt_gt = dict(list(enumerate([0] + loc.ALT))[1:])
        gt_dict.update(alt_gt)
        loc_s = str(loc).strip().split("\t")
        format_str = loc_s[8].strip().split(":")

        # Process genotype calls
        # gt calls
        gt_set = [dict(remove_missing_fields(
            list(zip(format_str, list(map(filter_format_null, x.split(":"))))))) for x in loc_s[9:]]

        # add sample names
        [x.update({"SAMPLE": sample, "TGT": format_tgt(x, gt_dict)})
         for sample, x in zip(v.samples, gt_set)]

        # Insert Data
        if args["--ANN"]:
            record_inserted = False # Ensure at least one record inserted
            for i in annotation_record_set:
                loc_set.append(copy.copy(i))
                record_inserted = True
            if not record_inserted:
                loc_set.append({})
        else:
            loc_set.append({})
        
        field_names = vcf_table._meta.sorted_field_names[1:]
        for rec in loc_set:
            rec["Variant"] = variant_index
            rec["CHROM"] = loc.CHROM
            rec["POS"] = loc.POS
            rec["_ID"] = loc.ID
            rec["REF"] = loc.REF
            rec["ALT"] = '|'.join(loc.ALT)
            rec["QUAL"] = loc.QUAL
            rec["FILTER"] = loc.FILTER
            if args["--compress"]:
                rec["GT"] = base64.b64encode(zlib.compress(pickle.dumps(gt_set).encode("utf-8")))
            else:
                rec["GT"] = json.dumps(gt_set)

            
            for k in field_names:
                if k not in rec:
                    rec[k] = ""
                elif rec[k] is None:
                    rec[k] = ""
            if args["--print"]:
                csv_writer.writerow(["0"] + [str(rec[x]) for x in field_names])
            else:
                insert_set.append(rec)
            loc_set = []
        if c % 15000 == 0 and not args["--print"]:
            with db.atomic():
                vcf_table.insert_many(insert_set).execute()
            print(("Inserted {c} records".format(c=c)))
            insert_set = []
    if not args["--print"]:
        with db.atomic():
                vcf_table.insert_many(insert_set).execute()

with indent(4):
    puts_err(colored.blue("\nDB Setup Complete\n"))



import re
import os
from subprocess import call
from clint.textui import colored, puts, puts_err, indent, progress

def command_out(message):
    """
        Used to print shell commands to stderr
    """
    puts_err(colored.green(message))
    

def message(message, n_indent = 4, color = "blue"):
    with indent(n_indent):
        if color == "blue":
            puts_err(colored.blue('\n' + message + '\n'))
        elif color == "red":
            puts_err(colored.red('\n' + message + '\n'))


def boolify(s):
    if s == 'True':
        return True
    if s == 'False':
        return False
    raise ValueError("huh?")

def autoconvert(s):
    for fn in (boolify, int, float):
        try:
            return fn(s)
        except ValueError:
            pass
    return s

def parse_region(region):
    return re.split("[:-]+", region)


# Stack Overflow: 377017
def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def check_program_exists(program):
    if which(program) is None:
        exit(puts_err(colored.red("\n\t" + program + " not installed or on PATH.\n")))


def run_command(comm, shell = True):
    """
        Runs a shell command
    """
    sh_out = ' '.join(comm) if type(comm) == list else comm
    command_out(sh_out)
    out = call(comm, shell = shell)
    if out != 0:
        raise Exception(f"Error [{out}] running {sh_out}")
    return out


# Levenshtein edit distnace
# https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
def lev(s1, s2):
    if len(s1) < len(s2):
        return lev(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = list(range(len(s2) + 1))
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1 # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1       # than s2
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    
    return previous_row[-1]




from setuptools import setup
import glob
import os
from setuptools import dist
dist.Distribution().fetch_build_eggs(['Cython>=0.24.1', 'numpy>=1.11'])

with open('requirements.txt') as f:
    required = f.read().splitlines()

from vcfkit import __version__

def gen_data_files(*dirs):
    results = []

    for src_dir in dirs:
        for root,dirs,files in os.walk(src_dir):
            results.append((root, [root + "/" + f for f in files]))
    return results


setup(name='VCF-kit',
      version=__version__,
      packages=['vcfkit','vcfkit.utils'],
      author='Daniel E. Cook',
      author_email='danielecook@gmail.com',
      license='MIT',
      entry_points="""
      [console_scripts]
      vk = vcfkit.vk:main
      """,
      install_requires=required,
      zip_safe=False,
      package_data={
        '': ['static/*', 'static/**/*'],
      },
      include_package_data=True,
      keywords=["VCF", "variant", "caller", "format", "snps", "genetic", "variation", "genetics"],
      data_files=gen_data_files("static"),
      setup_requires=['pytest-runner'],
      tests_require=['pytest', 'coveralls'],
      classifiers=["Development Status :: 4 - Beta","Operating System :: MacOS",
                   "Operating System :: Unix",
                   "Operating System :: POSIX :: Linux",
                   "License :: OSI Approved :: MIT License"])



#! /usr/bin/env python
"""
usage:
  vk vcf2tsv (wide|long) [--print-header --ANN] <vcf>

options:
  -h --help                   Show this screen.
  --version                   Show version.

"""
import os
import re
import sys
from signal import SIG_DFL, SIGPIPE, signal
from subprocess import PIPE, Popen

from docopt import docopt
from vcfkit.utils.vcf import *

signal(SIGPIPE, SIG_DFL)

# Info
r_info = re.compile(r'''\#\#INFO=<
  ID=(?P<id>[^,]+),
  Number=(?P<number>-?\d+|\.|[AG]),
  Type=(?P<type>Integer|Float|Flag|Character|String),
  Description="(?P<desc>[^"]*)".*
  >''', re.VERBOSE)

# Format
r_format = re.compile(r'''\#\#FORMAT=<
  ID=(?P<id>.+),
  Number=(?P<number>-?\d+|\.|[AG]),
  Type=(?P<type>.+),
  Description="(?P<desc>.*)".*
  >''', re.VERBOSE)

ANN_header = ["allele",
              "effect",
              "impact",
              "gene_name",
              "gene_id",
              "feature_type",
              "feature_id",
              "transcript_biotype",
              "exon_intron_rank",
              "nt_change",
              "aa_change",
              "cDNA_position/cDNA_len",
              "protein_position",
              "distance_to_feature",
              "error"]

debug = None
if len(sys.argv) == 1:
    debug = ['genome', "test.vcf.gz"]

if __name__ == '__main__':
    args = docopt(__doc__,
                  argv=debug,
                  options_first=False)
    module_path = os.path.split(os.path.realpath(__file__))[0]
    v = vcf(args["<vcf>"])
    field_set = []
    if len(sys.argv) == 1:
        print("Specify a VCF File")
        sys.exit()
    else:
        info = [m.groupdict()["id"] for m in r_info.finditer(v.raw_header)]
        format = [m.groupdict()["id"] for m in r_format.finditer(v.raw_header)]

    # Construct Query String
    print_header = ""
    if args["--print-header"]:
        print_header = "--print-header"
    if args["long"]:
        query_start = repr("%CHROM\t%POS\t%ID\t%REF\t%ALT\t%QUAL\t%FILTER\t" +
                           '\t'.join(['%' + x for x in info]) + "\t[-->%SAMPLE\t" +
                           '\t'.join(['%' + x for x in format]) + "\n]").strip("'")
    elif args["wide"]:
        query_start = repr("%CHROM\t%POS\t%ID\t%REF\t%ALT\t%QUAL\t%FILTER\t" +
                           '\t'.join(['%INFO/' + x for x in info]) +
                           "[\t%SAMPLE\t" + '\t'.join(['%' + x for x in format]) + "]\n").strip("'")
    comm = list(filter(len, ["bcftools", "query", print_header, "-f", query_start, v.filename]))
    comm = Popen(comm, stdout=PIPE, stderr=PIPE)
    if args["--ANN"]:
        ANN_loc = info.index("ANN") + 7
    for n, line in enumerate(comm.stdout):
        line = line.decode("UTF-8")
        if n == 0 and args["--print-header"] and args["wide"]:
            # Split out snpeff annotations
            if "ANN" in info and args["--ANN"]:
                line = line.split("\t")
                line = line[:ANN_loc - 1] + ANN_header + line[ANN_loc + 1:]
                line = '\t'.join(line)
            print(re.sub("\[[0-9]+\]", "", line).strip("#\n ").replace(":", "_"))
        elif n == 0 and args["long"] and args["--print-header"]:
            # Fix header for long format.
            line = re.sub("\[[0-9]+\]", "", line).strip("#\n ").split("\t")
            header = []
            for var in line:
                if ":" in var:
                    var = var.split(":")[1]
                header.append(var)
            if "ANN" in info and args["--ANN"]:
                header = header[:ANN_loc - 1] + ANN_header + header[ANN_loc + 1:]
            # Add Format to FORMAT
            header = header[:-len(format)] + ["F_" + x for x in header[-len(format):]]
            print(("\t".join(header).replace("-->", "")))
        elif n < len(v.samples) and args["long"]:
            pass
        elif n >= len(v.samples) and args["long"]:
            line = line.strip("\n").split("\t")
            if line[0].startswith("-->"):
                line = fill_fields + line[len(line) - len(fill_fields):]
            else:
                fill_fields = line[0:(7 + len(info))]
            if "ANN" in info and args["--ANN"]:
                for var_effect in line[ANN_loc].split(","):
                    out_line = line[:ANN_loc - 1] + var_effect.split("|") + line[ANN_loc + 2:]
                    print(('\t'.join(out_line).strip("\n").replace("-->", "")))
            else:
                print(("\t".join(line).replace("-->", "")))
        else:
            # Print wide format
            if "ANN" in info and args["--ANN"]:
                line = line.split("\t")
                for var_effect in line[ANN_loc].split(","):
                    out_line = line[:ANN_loc - 1] + var_effect.split("|") + line[ANN_loc + 2:]
                    print(('\t'.join(out_line).strip("\n")))
            else:
                print((line.strip("\n")))



#! /usr/bin/env python
"""
usage:
  vk filter (REF|HET|ALT|MISSING) [--min=<min> --max=<max> --soft-filter=<soft> --mode=(+|x)] <vcf>

Example

options:
  -h --help                   Show this screen.
  --version                   Show version.



"""
import os
import sys
from docopt import docopt
from clint.textui import colored, puts, indent
from vcfkit.utils.vcf import *
from vcfkit.utils.fasta import *
from vcfkit.utils import message
from collections import defaultdict
from glob import glob
from pprint import pprint as pp
from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE,SIG_DFL) 


def main(debug = None):
    args = docopt(__doc__, 
                  version='VCF-Toolbox v0.1',
                  argv = debug,
                  options_first=False)
    if args["--soft-filter"] and not args["--mode"]:
        exit(message("Must Specify --mode with soft-filter"))
    v = vcf(args["<vcf>"])
    n_samples = len(v.samples) * 1.0
    f = {}
    filter_s = [x for x in list(args.values()) if x in ["REF","HET","ALT","MISSING"]][0]
    # Filter by rate or by number?
    if args["--min"]:
        direction = "<"
        if int(float(args["--min"])) != float(args["--min"]):
            filter_key_min = "r_" + filter_s
            filter_val_min = float(args["--min"])
            filter_type = "FREQUENCY"
        else:
            filter_key_min = filter_s
            filter_val_min = int(float(args["--min"]))
            filter_type = "COUNT"
        filter_value = filter_val_min
    if args["--max"]:
        direction = ">"
        if int(float(args["--max"])) != float(args["--max"]):
            filter_key_max = "r_" + filter_s
            filter_val_max = float(args["--max"])
            filter_type = "FREQUENCY"
        else:
            filter_key_max = filter_s
            filter_val_max = int(float(args["--max"]))
            filter_type = "COUNT"
        filter_value = filter_val_max

    # Output header
    header = v.raw_header.splitlines()
    for n, i in enumerate(header):
        if i.startswith("##FILTER") and args["--soft-filter"]:
            filter_name = args["--soft-filter"]
            filter_line = """##FILTER=<ID={filter_name},Description="Apply filter if {filter_type}({filter_s}) {direction} {filter_value}">""".format(**locals())
            header.insert(n+1, filter_line)
            break
    header = '\n'.join(header) + "\n"
    sys.stdout.write(header)
    for line in v:
        filtered = False
        f["ALT"] = line.num_hom_alt
        f["HET"] = line.num_het
        f["REF"] = line.num_hom_ref
        f["MISSING"] = int(n_samples - line.num_called)
        f["r_ALT"] = f["ALT"] / n_samples
        f["r_HET"] = f["HET"] / n_samples
        f["r_REF"] = f["REF"] / n_samples
        f["r_MISSING"] = f["MISSING"] / n_samples
        if args["--min"]:
            if f[filter_key_min] < filter_val_min:
                filtered = True
        if args["--max"]:
            if f[filter_key_max] > filter_val_max:
                filtered = True
        if args["--soft-filter"]:
            line = str(line).split("\t")
            if args["--mode"] == "x":
                line[6] = "PASS"
            if filtered is False:
                sys.stdout.write('\t'.join(line))
            else:
                if args["--mode"] == "+":
                    if line[6] == "PASS":
                        line[6] = ""
                    line[6] = ';'.join([line[6]] + [args["--soft-filter"]]).strip(";")
                elif args["--mode"] == "x":
                    line[6] = args["--soft-filter"]
                sys.stdout.write('\t'.join(line))
        elif filtered is False:
            sys.stdout.write(str(line))

if __name__ == '__main__':
    main()


#! /usr/bin/env python
"""
usage:
  vk calc sample_hom_gt <vcf>
  vk calc genotypes [--frequency] <vcf>
  vk calc spectrum <vcf>

Example

options:
  -h --help                   Show this screen.
  --version                   Show version.

"""
from vcfkit import __version__
from docopt import docopt
from collections import defaultdict
from vcfkit.utils.vcf import *
from vcfkit.utils.fasta import *
from vcfkit.utils import autoconvert


class freq_vcf(vcf):

    """
        Subclass of vcf that calculates frequency of alleles by strain
    """

    def __init__(self, filename):
        vcf.__init__(self, filename)

    def calc_af(self, args):
        af_freq = {}
        af_freq = {sample: defaultdict(int) for sample in self.samples}
        for line in self:
            """
            0/0 -> 0
            0/1 -> 1
            ./. -> 2
            1/1 -> 3
            """
            for i in [sample for sample, gt in zip(self.samples, line.gt_types) if gt == 3]:
                af_freq[i][line.num_hom_alt] += 1
        # Output results
        print("\t".join(["sample", "freq_of_gt", "n_gt_at_freq"]))
        for sample in list(af_freq.keys()):
            for i in range(1, len(self.samples) + 1):
                out = "\t".join(map(str, [sample, i, af_freq[sample][i]]))
                print(out)

    def calc_genotypes(self, args):
        """
            Calculate count/frequency of genotypes.
        """
        s_freq = defaultdict(int)
        n_samples = len(self.samples)
        for line in self:
            ref = line.num_hom_ref
            alt = line.num_hom_alt
            het = line.num_het
            mis = line.num_unknown
            if args["--frequency"]:
                ref = float(ref) / n_samples
                het = float(het) / n_samples
                alt = float(alt) / n_samples
                mis = float(mis) / n_samples
            s_site = '__'.join(map(str, [ref, het, alt, mis]))
            s_freq[s_site] += 1
        freq_set = [(int(n), list(map(autoconvert, x.split("__")))) for n, x in zip(list(s_freq.values()), list(s_freq.keys()))]
        freq_set = sorted(freq_set, key=lambda s: s[0], reverse=True)
        print("n\tref\thet\talt\tmis")
        for n, freqs in freq_set:
            print((str(n) + '\t' + '\t'.join(map(str, freqs))))

    def calc_spectrum(self, args):
        """
            Calculate ALT Allele Frequency Spectrum
        """
        s_freq = defaultdict(int)
        print("n\talt_allele_freq")
        for line in self:
            s_freq[line.aaf] += 1
        s_freq = sorted(list(s_freq.items()), key=lambda s: s[0], reverse=True)
        for freq, n in s_freq:
            print((str(n) + "\t" + str(freq)))


def main(debug=None):
    # Define args globally
    args = docopt(__doc__,
                  argv=debug,
                  options_first=False,
                  version=__version__)
    vcf = freq_vcf(args["<vcf>"])
    if args["sample_hom_gt"]:
        vcf.calc_af(args)
    elif args["genotypes"]:
        vcf.calc_genotypes(args)
    elif args["spectrum"]:
        vcf.calc_spectrum(args)

if __name__ == '__main__':
    main()



import re
import os
import sys
from cyvcf2 import VCF as cyvcf2
from cyvcf2 import VCFReader
from clint.textui import colored, puts, indent
from collections import OrderedDict, defaultdict
from vcfkit.utils import lev, message
from copy import copy
import numpy as np
from vcfkit.utils.primer3 import primer3
from subprocess import Popen, PIPE, check_output
from .reference import resolve_reference_genome
np.set_printoptions(threshold=sys.maxsize)
from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE, SIG_DFL)

from Bio.Seq import Seq
from Bio.Restriction import AllEnzymes, CommOnly, RestrictionBatch

# Global flag for header output
header_printed = False

high_fidelity = RestrictionBatch(["AgeI",
                 "ApoI",
                 "BamHI",
                 "BbsI",
                 "BmtI",
                 "BsaI",
                 "BsiWI",
                 "BsrGI",
                 "BstEII",
                 "BstZ17I",
                 "DraIII",
                 "EagI",
                 "EcoRI",
                 "EcoRV",
                 "HindIII",
                 "KpnI",
                 "MfeI",
                 "MluI",
                 "NcoI",
                 "NheI",
                 "NotI",
                 "NruI",
                 "NsiI",
                 "PstI",
                 "PvuI",
                 "PvuII",
                 "SacI",
                 "SalI",
                 "SbfI",
                 "ScaI",
                 "SpeI",
                 "SphI",
                 "SspI",
                 "StyI"])

debug = None


class cvariant:
    """
        Mutable variant object
    """
    def __init__(self, variant):
        for i in [x for x in dir(variant) if x.startswith("_") is False]:
            setattr(self, i, getattr(variant, i))


class template:

    def __init__(self, variant, reference_file, vcf, use_template, nprimers):
        # Copy variant attributes over
        for i in [x for x in dir(variant) if x.startswith("_") is False]:
            setattr(self, i, getattr(variant, i))
        self.output_samples = vcf.output_samples
        self.reference_file = reference_file
        self.region_start = variant.POS - variant.region_size
        if self.region_start <= 0:
            self.region_start = 1
        self.region_end = variant.POS + variant.region_size - 1
        self.region = "{self.CHROM}:{self.region_start}-{self.region_end}".format(**locals())
        
        # Retrieve sequences
        self.ref_seq = self.fetch_sequence("REF")
        if use_template == 'REF':
            self.alt_seq = self.ref_seq
        else:
            self.alt_seq = self.fetch_sequence(use_template)
        self.edit_distance = lev(str(self.ref_seq).lower(), str(self.alt_seq).lower())

        # Retrieve primers
        if self.mode != 'template':
            p3 = primer3(reference_file)
            p3.PRIMER_NUM_RETURN = nprimers
            # Modify amplicon size if indel.
            if self.mode == 'indel':
                max_amplicon_length = max([len(self.ref_seq), len(self.alt_seq)])
                min_amplicon_length = max_amplicon_length - 200
                if (min_amplicon_length < 50):
                    min_amplicon_length = 50
                p3.PRIMER_PRODUCT_SIZE_RANGE = "{min_amplicon_length}-{max_amplicon_length}".format(**locals())
            elif self.mode == 'sanger':
                p3.PRIMER_PRODUCT_SIZE_RANGE = "{self.amplicon_lower}-{self.amplicon_upper}".format(**locals())
            primers = p3.fetch_primers(self.ref_seq, self.CHROM, self.region_start)
            if primers:
                self.primers = [x for x in primers if x.filter_primer_group()]
            else:
                self.primers = []

        # Identify restriction site locations
        if self.mode == 'snip':
            ref_seq = str(self.ref_seq)
            # Generate a primary variant only seq to identify restriction
            # sites that target ALT sites.
            self.primary_variant_seq = Seq(ref_seq[0:500] + self.ALT[0] + ref_seq[501:])
            self.fetch_restriction_sites(vcf.enzymes)


    def fetch_sequence(self, use_template):
        """
            Fetches sequence surrounding variant for REF, ALT, or given sample.
        """
        sample_flag = ""
        if use_template == "REF":
            command = f"samtools faidx {self.reference_file} {self.region}"
        elif use_template == "ALT" or use_template is None:
            command = f"samtools faidx {self.reference_file} {self.region} | bcftools consensus {self.filename}"
        else:
            sample_flag = "--sample=" + use_template  # Get use_template for sample.
            command = f"samtools faidx {self.reference_file} {self.region} | bcftools consensus {sample_flag} {self.filename}"
        try:
            seq = check_output(command, shell=True)
            seq = Seq(''.join(seq.decode('utf-8').splitlines()[1:]))
        except:
            seq = Seq('')
        return seq


    def fetch_restriction_sites(self, enzymes = "Common"):
        """ 
            Spike in target variant first, generate list 
            restriction enzymes that will work.
        """
        if enzymes == "ALL":
            enzyme_group = AllEnzymes
        elif enzymes == "Common":
            enzyme_group = CommOnly
        elif enzymes == "HF":
            enzyme_group = high_fidelity
        else:
            enzyme_group = RestrictionBatch(enzymes.split(","))

        # Filter ambiguous cutters
        enzyme_group = RestrictionBatch([x for x in enzyme_group if x.is_ambiguous() is False])

        # Calculate rflps for ALT sites only
        self.ref_sites = dict(list(enzyme_group.search(self.ref_seq).items()))
        self.primary_variant_sites = dict(list(enzyme_group.search(self.primary_variant_seq).items()))
        self.rflps = {k: (self.ref_sites[k], self.primary_variant_sites[k]) for k, v in list(self.ref_sites.items()) 
                        if len(v) > 0 and len(v) <= 3
                        and self.ref_sites[k] != self.primary_variant_sites[k]}


    def calculate_cuts(self, length, positions):
        """
            Given a template and cut sites,
            returns resulting fragments and their size.
        """
        if positions:
            init, cuts = 0, []
            for cut in positions:
                cut_len = cut - init
                cuts.append(cut_len)
                init = init + cut_len
            # Add last piece in
            cuts.append(length - positions[-1])
            return cuts
        else:
            return [length]


    def fetch_variant_count(self, region, sample_set = None):
        """
            Fetches variant count in a region.
        """
        v = VCFReader(self.filename)
        sample_indices = [v.samples.index(x) for x in sample_set]
        vc = 0
        for i in v(self.region):
            gt_len = len(set([i.gt_types[x] for x in sample_indices]))
            if gt_len > 1:
                vc += 1
        return vc



    def print_header(self, hout):
        global header_printed
        if header_printed is False:
            print(('\t'.join(map(str, hout))))
            header_printed = True

    def out(self):

        hout = ["CHROM",
                "POS",
                "region",
                "REF",
                "ALT",
                "template_sample"]

        out = [self.CHROM,
               self.POS,
               self.region,
               self.REF,
               ','.join(self.ALT),
               self.use_template]
        homozygous_ref = ','.join(map(str,self.gt_collection[0]))
        homozygous_alt = ','.join(map(str,self.gt_collection[3]))

        if self.mode == 'snip':
            hout += ["variant_count",
                     "ref_sites",
                     "alt_sites",
                     "restriction_enzyme",
                     "restriction_site",
                     "restriction_site_length",
                     "primer_left",
                     "primer_right",
                     "melting_temperature",
                     "amplicon_length",
                     "amplicon_region",
                     "amplicon_sequence",
                     "0/0",
                     "1/1",
                     "polymorphic"]

            self.print_header(hout)
            for primer_group in self.primers:
                amp_start = self.region_start + primer_group.primer_left.START
                amp_end = self.region_start + primer_group.primer_right.END
                amplicon_length = len(primer_group.amplicon)
                primer_group.amplicon_region = "{self.CHROM}:{amp_start}-{amp_end}".format(**locals())
                if primer_group is not None:
                    for rflp, cut_sites in list(self.rflps.items()):
                        # Shift cut sites
                        ref_cut_sites = [x - primer_group.primer_left.START for x in cut_sites[0]]
                        alt_cut_sites = [x - primer_group.primer_left.START for x in cut_sites[1]]
                        # Calculate Cuts and shift positions
                        ref_product_sizes = self.calculate_cuts(amplicon_length, ref_cut_sites)
                        alt_product_sizes = self.calculate_cuts(amplicon_length, alt_cut_sites)

                        # FILTER: If product sizes are less than 25, ignore.
                        product_sizes = ref_product_sizes + alt_product_sizes
                        if any([x < 25 for x in product_sizes]):
                            break

                        # FILTER: Remove sites where the number of restriction sites does not change.
                        if len(ref_cut_sites) == len(alt_cut_sites):
                            break

                        ref_cut_sites = ','.join(map(str, ref_cut_sites))
                        ref_product_sizes = ','.join(map(str, ref_product_sizes))
                        alt_cut_sites = ','.join(map(str, alt_cut_sites))
                        alt_product_sizes = ','.join(map(str, alt_product_sizes))

                        # Fetch variant Count
                        variant_count = self.fetch_variant_count(primer_group.amplicon_region, self.output_samples)

                        # out_primers represents variables that are specific to the primer/rflp
                        out_primers = out + [variant_count,
                                             ref_cut_sites + ":" + ref_product_sizes,
                                             alt_cut_sites + ":" + alt_product_sizes,
                                             rflp,
                                             rflp.site,
                                             rflp.size,
                                             primer_group.primer_left,
                                             primer_group.primer_right,
                                             primer_group.primer_tm,
                                             primer_group.amplicon_length,
                                             primer_group.amplicon_region,
                                             primer_group.amplicon,
                                             homozygous_ref,
                                             homozygous_alt,
                                             self.is_polymorphic]
                        print(('\t'.join(map(str, out_primers))))
        elif self.mode == 'indel':
            hout += ["variant_count",
                     "indel_size",
                     "indel_type",
                     "primer_left",
                     "primer_right",
                     "melting_temperature",
                     "REF_product_size",
                     "ALT_product_size",
                     "amplicon_region",
                     "amplicon_sequence",
                     "0/0",
                     "1/1",
                     "polymorphic"]

            self.print_header(hout)
            for primer_group in self.primers:
                # Fetch variant Count
                variant_count = self.fetch_variant_count(primer_group.amplicon_region, self.output_samples)
                out_indel = out + [variant_count,
                                   self.indel_size,
                                   self.indel_type,
                                   primer_group.primer_left,
                                   primer_group.primer_right,
                                   primer_group.primer_tm,
                                   primer_group.amplicon_length,
                                   primer_group.amplicon_length - self.indel_size,
                                   primer_group.amplicon_region,
                                   primer_group.amplicon,
                                   homozygous_ref,
                                   homozygous_alt,
                                   self.is_polymorphic]
                print(('\t'.join(map(str, out_indel))))

        elif self.mode == 'sanger':
            hout += ["variant_count",
                     "variant_distance",
                     "primer_left",
                     "primer_right",
                     "melting_temperature",
                     "amplicon_length",
                     "amplicon_region",
                     "amplicon_sequence",
                     "0/0",
                     "1/1",
                     "polymorphic"]
            self.print_header(hout)
            for primer_group in self.primers:
                amp_start = self.region_start + primer_group.primer_left.START
                # Fetch variant Count
                variant_count = self.fetch_variant_count(primer_group.amplicon_region, self.output_samples)
                out_indel = out + [variant_count,
                                   self.POS - amp_start,
                                   primer_group.primer_left,
                                   primer_group.primer_right,
                                   primer_group.primer_tm,
                                   primer_group.amplicon_length,
                                   primer_group.amplicon_region,
                                   primer_group.amplicon,
                                   homozygous_ref,
                                   homozygous_alt,
                                   self.is_polymorphic]
                print(('\t'.join(map(str, out_indel))))

        else:
            # Output template
            hout += ["template"]
            out += [self.alt_seq]
            self.print_header(hout)
            print(('\t'.join(map(str, out))))


class primer_vcf(cyvcf2):
    def __init__(self, filename, reference=None, use_template="ALT", polymorphic=True):
        if not filename.startswith("http") and not os.path.isfile(filename) and filename != "-":
            exit(message(f"Error: {filename} does not exist", color='red'))
        self.filename = filename
        self.use_template = use_template
        self.use_polymorphic = polymorphic
        if reference:
            self.reference = reference
            self.reference_file = resolve_reference_genome(reference)

        cyvcf2.__init__(self, self.filename)
        # Check if file exists
        self.n = len(self.samples)  # Number of Samples

        # Meta Data
        comp = re.compile(r'''^##(?P<key>[^<#]+?)=(?P<val>[^<#]+$)''', re.M)
        self.metadata = OrderedDict(comp.findall(self.raw_header))

        # Contigs
        self.contigs = OrderedDict(list(zip(
            re.compile("##contig=<ID=(.*?),").findall(self.raw_header),
            list(map(int, re.compile("##contig.*length=(.*?)>").findall(self.raw_header)))
        )))

        self.info_set = [x for x in self.header_iter() if x.type == "INFO"]
        self.filter_set = [x for x in self.header_iter() if x.type == "FILTER"]
        self.format_set = [x for x in self.header_iter() if x.type == "FORMAT"]
        self.header = copy(self.raw_header)

        # Make sure index exists for this VCF.
        if not self.filename.startswith("http") and not os.path.isfile(self.filename + ".csi"):
            message("VCF is not indexed; Indexing.")
            comm = "bcftools index {f}".format(f=self.filename)
            out = check_output(comm, shell=True)
            message(out.decode("utf-8"))

        # Setup use_template
        if self.use_template not in self.samples + ["REF", "ALT"]:
            exit(message(self.use_template + " use_template not found."))


    def variant_iterator(self):
        """
            Iterate over variants in region is specified;
            Otherwise start at the beginning.
        """
        if self.region:
            var_region = self.fetch_variants(self.region)
        else:
            var_region = self

        for variant in var_region:

            # Skip variants that are not applicable
            if self.mode == 'indel':
                if variant.is_indel is False:
                    continue
            elif self.mode == 'snip':
                if variant.is_snp is False:
                    continue

            nz = cvariant(variant)

            # Process Indels
            if nz.is_indel:
                # Size is dynamically set for indels
                nz.indel_size = len(nz.REF) - len(nz.ALT[0])
                nz.indel_size_abs = abs(len(nz.ALT[0]) - len(nz.REF))
                # Indel type:
                if nz.indel_size > 0:
                    nz.indel_type = "deletion"
                else:
                    nz.indel_type = "insertion"

            if self.mode == 'snip':
                nz.region_size = 500
            elif self.mode == 'indel':
                nz.region_size = nz.indel_size * 2 + 150
                if nz.indel_size_abs <= 25:
                    continue
            elif self.mode in ['sanger', 'template']:
                nz.region_size = self.region_size
            nz.use_template = self.use_template
            nz.mode = self.mode
            nz.filename = self.filename
            nz.use_polymorphic = self.use_polymorphic

            gt_collection = defaultdict(list)

            for vcf_sample, gt in zip(self.samples, variant.gt_types):
                if vcf_sample in self.output_samples:
                    gt_collection[gt].append(vcf_sample)

            if gt_collection:
                nz.gt_collection = gt_collection
                nz.homozygous_ref = nz.gt_collection[0]
                nz.heterozygous = nz.gt_collection[1]
                nz.homozygous_alt = nz.gt_collection[3]
                nz.missing = nz.gt_collection[2]
            is_polymorphic = all([len(x) > 0 for x in [gt_collection[0], gt_collection[3]]])

            nz.is_polymorphic = is_polymorphic
            if self.use_polymorphic and not is_polymorphic:
                continue

            if self.mode == 'sanger':
                nz.amplicon_lower = self.amplicon_lower
                nz.amplicon_upper = self.amplicon_upper

            t = template(nz, self.reference_file, self, self.use_template, self.nprimers)
            yield t

    def fetch_variants(self, region=None):
        """
            Fetch variants using the specified region
        """
        if region is None:
            region = self.region
        for i in self(region):
            yield i

